import logging
import atexit

import rclpy

from wrapyfi.utils import SingletonOptimized
from wrapyfi.connect.wrapper import MiddlewareCommunicator


class ROS2Middleware(metaclass=SingletonOptimized):
    """
    ROS 2 middleware wrapper. This class is a singleton, so it can be instantiated only once. The ``activate`` method
    should be called to initialize the middleware. The ``deinit`` method should be called to deinitialize the middleware
    and destroy all connections. The ``activate`` and ``deinit`` methods are automatically called when the class is
    instantiated and when the program exits, respectively.
    """

    @staticmethod
    def activate(**kwargs):
        """
        Activate the ROS 2 middleware. This method should be called to initialize the middleware.

        :param kwargs: dict: Keyword arguments to be passed to the ROS 2 initialization function
        """
        ROS2Middleware(**kwargs)

    def __init__(self, *args, **kwargs):
        """
        Initialize the ROS 2 middleware. This method is automatically called when the class is instantiated.

        :param args: list: Positional arguments to be passed to the ROS 2 initialization function
        :param kwargs: dict: Keyword arguments to be passed to the ROS 2 initialization function
        """
        logging.info("Initialising ROS 2 middleware")
        rclpy.init(args=[*args], **kwargs)
        atexit.register(MiddlewareCommunicator.close_all_instances)
        atexit.register(self.deinit)

    @staticmethod
    def deinit():
        """
        Deinitialize the ROS 2 middleware. This method is automatically called when the program exits.
        """
        logging.info("Deinitializing ROS 2 middleware")
        if rclpy.ok():
            rclpy.shutdown()
        else:
            logging.info(
                "ROS 2 context is already shutdown or not initialized. Skipping shutdown."
            )



"""
Encoder and Decoder for MXNet Tensor Data via Wrapyfi.

This script provides mechanisms to encode and decode MXNet tensor data using Wrapyfi.
It leverages base64 encoding to convert binary data into ASCII strings.

The script contains a class, `MXNetTensor`, registered as a plugin to manage the
conversion of MXNet tensor data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - MXNet: A deep learning framework designed for both efficiency and flexibility (refer to https://mxnet.apache.org/get_started for installation instructions)
        Note: If MXNet is not available, HAVE_MXNET will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install mxnet``  # Basic installation of MXNet
"""

import io
import base64
from functools import lru_cache

import numpy as np

from wrapyfi.utils import *


try:
    import mxnet

    HAVE_MXNET = True
except ImportError:
    HAVE_MXNET = False


def mxnet_device_to_str(device):
    """
    Convert an MXNet device to a string representation.

    :param device: Union[str, mxnet.Context, dict]: Various possible types representing an MXNet device as mxnet.Context or str. Also accepts a dictionary mapping encoded device strings to decoding devices
    :return: Union[str, dict]: A string or dictionary representing the MXNet device
    """
    if device is None:
        return "cpu:0"
    elif isinstance(device, dict):
        device_rets = {}
        for k, v in device.items():
            device_rets[mxnet_device_to_str(k)] = mxnet_device_to_str(v)
        return device_rets
    elif isinstance(device, mxnet.Context):
        return f'{device.device_type.replace("gpu", "cuda")}:{device.device_id}'
    elif isinstance(device, str):
        return device.replace("gpu", "cuda")
    else:
        raise ValueError(f"Unknown device type {type(device)}")


@lru_cache(maxsize=None)
def mxnet_str_to_device(device):
    """
    Convert a string to an MXNet device.

    :param device: str: A string representing an MXNet device
    :return: mxnet.Context: An MXNet context representing the device
    """
    if device is None:
        return mxnet.cpu()
    elif isinstance(device, mxnet.Context):
        return device
    elif isinstance(device, str):
        try:
            device_type, device_id = device.split(":")
        except ValueError:
            device_type = device
            device_id = 0
        return mxnet.Context(device_type.replace("cuda", "gpu"), int(device_id))
    else:
        raise ValueError(f"Unknown device type {type(device)}")


@PluginRegistrar.register(
    types=None if not HAVE_MXNET else mxnet.nd.NDArray.__mro__[:-1]
)
class MXNetTensor(Plugin):
    def __init__(self, load_mxnet_device=None, map_mxnet_devices=None, **kwargs):
        """
        Initialize the MXNetTensor plugin.

        :param load_mxnet_device: Union[mxnet.Context, str]: Default MXNet device to load tensors onto
        :param map_mxnet_devices: dict: [Optional] A dictionary mapping encoded device strings to decoding devices
        """
        self.map_mxnet_devices = map_mxnet_devices or {}
        if load_mxnet_device is not None:
            self.map_mxnet_devices["default"] = load_mxnet_device
        self.map_mxnet_devices = mxnet_device_to_str(self.map_mxnet_devices)

    def encode(self, obj, *args, **kwargs):
        """
        Encode MXNet tensor data into a base64 ASCII string.

        :param obj: mxnet.nd.NDArray: The MXNet tensor data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, encoded data string, and device string.
        """
        with io.BytesIO() as memfile:
            np.save(memfile, obj.asnumpy())
            obj_data = base64.b64encode(memfile.getvalue()).decode("ascii")
        obj_device = mxnet_device_to_str(obj.context)
        return True, dict(
            __wrapyfi__=(str(self.__class__.__name__), obj_data, obj_device)
        )

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into MXNet tensor data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string and device string
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, mxnet.nd.NDArray]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - mxnet.nd.NDArray: The decoded MXNet tensor data
        """
        with io.BytesIO(base64.b64decode(obj_full[1].encode("ascii"))) as memfile:
            obj_device = self.map_mxnet_devices.get(
                obj_full[2], self.map_mxnet_devices.get("default", None)
            )
            if obj_device is not None:
                obj_device = mxnet_str_to_device(obj_device)
                return True, mxnet.nd.array(np.load(memfile), ctx=obj_device)
            else:
                return True, mxnet.nd.array(np.load(memfile))



"""
A message publisher and listener for native Python objects and TensorFlow tensors.

This script demonstrates the capability to transmit native Python objects and TensorFlow tensors using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a nested dummy Python object with native objects and TensorFlow tensors
    - Applying the PUB/SUB pattern with mirroring

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - TensorFlow: Used for creating and handling tensors (refer to https://www.tensorflow.org/install for installation instructions)

    Install using pip:
        ``pip install "tensorflow>=2.9.1"``  # Basic installation of TensorFlow

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits message
        ``python3 tensorflow_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and prints the entire dummy object
        ``python3 tensorflow_example.py --mode listen``
"""

import argparse

try:
    import tensorflow as tf
except ImportError:
    print("Install TensorFlow before running this script.")

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "Notifier",
        "/notify/test_native_exchange",
        carrier="",
        should_wait=True,
    )
    def exchange_object(self, mware=None):
        """
        Exchange messages with TensorFlow tensors and other native Python objects.
        """
        msg = input("Type your message: ")

        ret = {
            "message": msg,
            "tf_ones": tf.ones((2, 4)),
            "tf_string": tf.constant("This is string"),
        }
        return (ret,)


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="A message publisher and listener for native Python objects and TensorFlow tensors."
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="publish",
        choices={"publish", "listen"},
        help="The transmission mode",
    )
    parser.add_argument(
        "--mware",
        type=str,
        default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission",
    )
    return parser.parse_args()


def main(args):
    """
    Main function to initiate Notifier class and communication.
    """
    notifier = Notifier()
    notifier.activate_communication(Notifier.exchange_object, mode=args.mode)

    while True:
        (msg_object,) = notifier.exchange_object(mware=args.mware)
        print("Method result:", msg_object)


if __name__ == "__main__":
    args = parse_args()
    main(args)



"""
Encoder and Decoder for CuPy Array Data via Wrapyfi.

This script provides mechanisms to encode and decode CuPy array data using Wrapyfi.
It leverages base64 encoding to convert binary data into ASCII strings.

The script contains a class, CuPyArray, registered as a plugin to manage the
conversion of CuPy array data (if available) between its original and encoded forms. CuPy only supports GPU devices,
specifically NVIDIA CUDA devices. Installing CUDA Toolkit (https://developer.nvidia.com/cuda-toolkit) is required
to use CuPy.

Requirements:
- Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
- CuPy: A GPU-accelerated library for numerical computations with a NumPy-compatible interface (refer to https://docs.cupy.dev/en/stable/install.html for installation instructions)
Note: If CuPy is not available, HAVE_CUPY will be set to False and the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install cupy-cuda12x``  # Basic installation of CuPy. Replace 12x with your CUDA version e.g., cupy-cuda11x
"""

import io
import base64

import numpy as np

from wrapyfi.utils import *

try:
    import cupy

    HAVE_CUPY = True
except ImportError:
    HAVE_CUPY = False


def cupy_device_to_str(device):
    """
    Convert a CuPy device to a string representation.

    :param device: Union[cupy.cuda.Device, int]: The CuPy device
    :return: str: A string representing the CuPy device
    """
    if isinstance(device, (cupy.cuda.Device, int)):
        return f"cuda:{int(device)}"
    elif isinstance(device, str):
        return device
    else:
        raise ValueError(f"Unknown device type {type(device)}")


def cupy_str_to_device(device_str):
    """
    Convert a string to a CuPy device.

    :param device_str: str: A string representing a CuPy device
    :return: cupy.cuda.Device: A CuPy device
    """
    if isinstance(device_str, str) and device_str.startswith("cuda:"):
        device_id = int(device_str.split(":")[1])
        return cupy.cuda.Device(device_id)
    elif isinstance(device_str, str) and device_str.startswith("cpu:"):
        raise ValueError("CuPY does not support CPU devices")
    else:
        raise ValueError(f"Invalid device string {device_str}")


@PluginRegistrar.register(types=None if not HAVE_CUPY else cupy.ndarray.__mro__[:-1])
class CuPyArray(Plugin):
    def __init__(self, load_cupy_device=None, map_cupy_devices=None, **kwargs):
        """
        Initialize the CuPy plugin.

        :param load_cupy_device: cupy.cuda.Device or str: Default CuPy device to load tensors onto
        :param map_cupy_devices: dict: [Optional] A dictionary mapping encoded device strings to decoding devices
        """
        self.map_cupy_devices = map_cupy_devices or {}
        if load_cupy_device is not None:
            self.map_cupy_devices["default"] = load_cupy_device
        self.map_cupy_devices = {
            k: cupy_device_to_str(v) for k, v in self.map_cupy_devices.items()
        }

    def encode(self, obj, *args, **kwargs):
        """
        Encode CuPy array into a base64 ASCII string.

        :param obj: cupy.ndarray: The CuPy array to encode
        :return: Tuple[bool, dict]
        """
        with io.BytesIO() as memfile:
            np.save(memfile, cupy.asnumpy(obj))
            obj_data = base64.b64encode(memfile.getvalue()).decode("ascii")
        obj_device = cupy_device_to_str(obj.device)
        return True, dict(
            __wrapyfi__=(str(self.__class__.__name__), obj_data, obj_device)
        )

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into CuPy array.

        :param obj_full: tuple: A tuple containing the encoded data string and device string
        :return: Tuple[bool, cupy.ndarray]
        """
        with io.BytesIO(base64.b64decode(obj_full[1].encode("ascii"))) as memfile:
            obj_device_str = self.map_cupy_devices.get(
                obj_full[2], self.map_cupy_devices.get("default", "cuda:0")
            )
            obj_device = cupy_str_to_device(obj_device_str)
            with obj_device:
                return True, cupy.array(np.load(memfile))



"""
Encoder and Decoder for Zarr Array/Group Data via Wrapyfi.

This script provides mechanisms to encode and decode Zarr Data using Wrapyfi.
It utilizes base64 encoding and zip compression.

The script contains a class, `ZarrData`, registered as a plugin to manage the
conversion of Zarr Data (Array and Group) (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - Zarr: A format for the storage of chunked, compressed, N-dimensional arrays (refer to https://zarr.readthedocs.io/en/stable/ for installation instructions)
        Note: If Zarr is not available, HAVE_ZARR will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install zarr``  # Basic installation of Zarr
"""

import tempfile
import io
import base64

from wrapyfi.utils import *

try:
    import zipfile
    import zarr

    HAVE_ZARR = True
except ImportError:
    HAVE_ZARR = False


@PluginRegistrar.register(
    types=None if not HAVE_ZARR else zarr.Array.__mro__[:-1] + zarr.Group.__mro__[:-1]
)
class ZarrData(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the ZarrData plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode Zarr Data using zip compression and base64.

        :param obj: Union[zarr.Array, zarr.Group]: The Zarr Data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, encoded data string, data type, and object name.
        """
        obj_type = "Group" if isinstance(obj, zarr.Group) else "Array"
        obj_name = obj.name if isinstance(obj, zarr.Array) else None

        with tempfile.TemporaryDirectory() as tmpdirname:
            store_path = os.path.join(tmpdirname, "zarr_store")

            if obj_type == "Array":
                zarr.save_array(store_path, obj)
            else:
                zarr.save_group(store_path, obj)

            with io.BytesIO() as binary_stream:
                with zipfile.ZipFile(binary_stream, "w") as zipf:
                    for foldername, subfolders, filenames in os.walk(store_path):
                        for filename in filenames:
                            zipf.write(
                                os.path.join(foldername, filename),
                                arcname=os.path.relpath(
                                    os.path.join(foldername, filename), store_path
                                ),
                            )
                obj_data = base64.b64encode(binary_stream.getvalue()).decode("ascii")

        return True, dict(
            __wrapyfi__=(str(self.__class__.__name__), obj_data, obj_type, obj_name)
        )

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a zip compressed and base64 encoded string back into Zarr Data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string, data type, and object name
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, Union[zarr.Array, zarr.Group]]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - Union[zarr.Array, zarr.Group]: The decoded Zarr Data
        """
        obj_data = obj_full[1]
        zarr_type = obj_full[2]
        zarr_name = obj_full[
            3
        ]  # zarr_name is used only if zarr_type is 'Array'. Currently not used.

        with io.BytesIO(base64.b64decode(obj_data.encode("ascii"))) as binary_stream:
            with tempfile.TemporaryDirectory() as tmpdirname:
                store_path = os.path.join(tmpdirname, "zarr_store")
                with zipfile.ZipFile(binary_stream, "r") as zipf:
                    zipf.extractall(path=store_path)

                if zarr_type == "Array":
                    array = zarr.open_array(store_path, mode="r")
                    return True, array
                else:
                    group = zarr.open_group(store_path, mode="r")
                    return True, group



import logging

from wrapyfi.connect.listeners import Listener, Listeners, ListenerWatchDog


@Listeners.register("MMO", "fallback")
class FallbackListener(Listener):

    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: str = "tcp",
        should_wait: bool = True,
        missing_middleware_object: str = "",
        **kwargs,
    ):
        logging.warning(
            f"Fallback listener employed due to missing middleware or object type: "
            f"{missing_middleware_object}"
        )
        Listener.__init__(
            self, name, in_topic, carrier=carrier, should_wait=should_wait, **kwargs
        )
        self.missing_middleware_object = missing_middleware_object

    def establish(self, repeats: int = -1, **kwargs):
        return None

    def listen(self):
        return None

    def close(self):
        return None



"""
A message publisher and listener for native Python objects and CuPy arrays.

This script demonstrates the capability to transmit native Python objects and CuPy arrays using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a nested dummy Python object with native objects and CuPy arrays
    - Applying the PUB/SUB pattern with mirroring
    - Transmitting CuPy arrays across different devices (only GPU devices are supported)

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - CuPy: Used for handling and creating arrays with GPU acceleration (refer to https://docs.cupy.dev/en/stable/install.html for installation instructions)

    Install using pip:
        ``pip install cupy-cuda12x``  # Basic installation of CuPy. Replace 12x with your CUDA version e.g., cupy-cuda11x

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits message
        ``python3 cupy_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and prints the entire dummy object
        ``python3 cupy_example.py --mode listen``
"""

import argparse

try:
    import cupy as cp
except ImportError:
    print("Install CuPy before running this script.")

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class CuPyNotifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "CuPyNotifier",
        "/notify/test_cupy_exchange",
        carrier="",
        should_wait=True,
        listener_kwargs=dict(load_cupy_device=0),  # cp.cuda.Device(0)
    )
    def exchange_object(self, mware=None):
        """
        Exchange messages with CuPy tensors and other native Python objects.
        """
        msg = input("Type your message: ")
        ret = {
            "message": msg,
            "cupy_ones_cuda": cp.ones((2, 4), dtype=cp.float32),
            "cupy_zeros_cuda": cp.zeros((2, 3), dtype=cp.float32),
        }
        return (ret,)


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="A message publisher and listener for native Python objects and CuPy tensors."
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="publish",
        choices={"publish", "listen"},
        help="The transmission mode",
    )
    parser.add_argument(
        "--mware",
        type=str,
        default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission",
    )
    return parser.parse_args()


def main(args):
    """
    Main function to initiate CuPyNotifier class and communication.
    """
    notifier = CuPyNotifier()
    notifier.activate_communication(CuPyNotifier.exchange_object, mode=args.mode)

    while True:
        (msg_object,) = notifier.exchange_object(mware=args.mware)
        print("Method result:", msg_object)


if __name__ == "__main__":
    args = parse_args()
    main(args)



"""
Encoder and Decoder for pandas Series/Dataframes Data via Wrapyfi.

This script provides mechanisms to encode and decode pandas data using Wrapyfi.
It utilizes base64 encoding to convert binary data into ASCII strings for "pandas<2.0".
It utilizes base64 encoding and pickle serialization for "pandas>=2.0".

The script contains a class, `PandasData`, registered as a plugin to manage the
conversion of pandas data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - pandas: A data structures library for data analysis, time series, and statistics (refer to https://pandas.pydata.org/pandas-docs/stable/getting_started/install.html for installation instructions)
        Note: If pandas is not available, HAVE_PANDAS will be set to False and
        the plugin will be registered with no types.
        For pandas>=2.0, the object is pickled (slower but identical on decoding).
        For pandas<2.0, the object is encoded using base64 (faster but potentially non-identical on decoding).

    You can install the necessary packages using pip:
        ``pip install pandas``  # Basic installation of pandas
"""

import io
import pickle
import base64

import numpy as np

from wrapyfi.utils import *

try:
    import pandas

    HAVE_PANDAS = True
except ImportError:
    HAVE_PANDAS = False


@PluginRegistrar.register(
    types=(
        None
        if not HAVE_PANDAS or (HAVE_PANDAS and pandas.__version__ < "2.0")
        else pandas.DataFrame.__mro__[:-1] + pandas.Series.__mro__[:-1]
    )
)
class PandasData(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the PandasData plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode pandas data using pickle and base64.

        :param obj: Union[pandas.DataFrame, pandas.Series]: The pandas data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, pickled data string, and any buffer data
        """
        buffers = []
        obj_data = pickle.dumps(obj, protocol=5, buffer_callback=buffers.append)
        obj_buffers = list(
            map(lambda x: base64.b64encode(memoryview(x)).decode("ascii"), buffers)
        )
        return True, dict(
            __wrapyfi__=(
                str(self.__class__.__name__),
                obj_data.decode("latin1"),
                *obj_buffers,
            )
        )

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a pickled and base64 encoded string back into pandas data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the pickled data string and any buffer data
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, Union[pandas.DataFrame, pandas.Series]]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - Union[pandas.DataFrame, pandas.Series]: The decoded pandas data
        """
        obj_data = obj_full[1].encode("latin1")
        obj_buffers = list(
            map(lambda x: base64.b64decode(x.encode("ascii")), obj_full[2:])
        )
        obj_data = bytearray(obj_data)
        for buf in obj_buffers:
            obj_data += buf
        return True, pickle.loads(obj_data, buffers=obj_buffers)


@PluginRegistrar.register(
    types=(
        None
        if not HAVE_PANDAS or (HAVE_PANDAS and pandas.__version__ >= "2.0")
        else pandas.DataFrame.__mro__[:-1] + pandas.Series.__mro__[:-1]
    )
)
class PandasLegacyData(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the PandasLegacyData plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode pandas data into a base64 ASCII string.

        :param obj: Union[pandas.DataFrame, pandas.Series]: The pandas data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, encoded data string, and object type
        """
        with io.BytesIO() as memfile:
            if isinstance(obj, pandas.DataFrame):
                obj_type = "DataFrame"
                obj.to_json(memfile, orient="split")
            elif isinstance(obj, pandas.Series):
                obj_type = "Series"
                obj.to_frame().to_json(memfile, orient="split")
            obj_data = base64.b64encode(memfile.getvalue()).decode("ascii")
        return True, dict(
            __wrapyfi__=(str(self.__class__.__name__), obj_data, obj_type)
        )

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into pandas data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string and object type
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, Union[pandas.DataFrame, pandas.Series]]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - Union[pandas.DataFrame, pandas.Series]: The decoded pandas data
        """
        with io.BytesIO(base64.b64decode(obj_full[1].encode("ascii"))) as memfile:
            obj = pandas.read_json(memfile, orient="split")
            obj_type = obj_full[2]
            if obj_type == "Series":
                obj = obj.iloc[:, 0]
        return True, obj



"""
A message publisher and listener for native Python objects, NumPy Arrays, and pandas Series/Dataframes.

This script demonstrates the capability to transmit native Python objects, NumPy arrays, and pandas series/dataframes using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a nested dummy Python object with native objects, NumPy arrays, and pandas series/dataframes
    - Applying the PUB/SUB pattern with mirroring

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - NumPy: Used for creating arrays (installed with Wrapyfi)
    - pandas: Used for creating and handling series and dataframes
    - PyArrow: Used for compatibility with pandas>=2.0. It is not required by pandas but required for this example when installing pandas >= 2.0. Not required for pandas < 2.0.

    Install using pip:
        ``pip install "pandas<2.0"``  # Basic installation of pandas<2.0 with Numpy as a dependency (for compatibility)
        ################## OR ##################
        ``pip install "pandas>=2.0" pyarrow``  # Basic installation of pandas>=2.0 with PyArrow as a dependency (for compatibility)

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits message
        ``python3 numpy_pandas_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and prints the entire dummy object
        ``python3 numpy_pandas_example.py --mode listen``
"""

import argparse

try:
    import numpy as np
    import pandas as pd
except ImportError:
    print("Install pandas and NumPy before running this script.")

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "Notifier",
        "/notify/test_native_exchange",
        carrier="tcp",
        should_wait=True,
    )
    def exchange_object(self, mware=None):
        """
        Exchange messages with NumPy arrays, pandas series/dataframes, and other native Python objects.
        """
        msg = input("Type your message: ")

        ret = {
            "message": msg,
            "numpy_array": np.ones((2, 4)),
            "pandas_series": pd.Series(
                [1, 3, 5, np.nan, 6, 8],
                dtype="float32" if pd.__version__ < "2.0" else "float32[pyarrow]",
            ),
            "pandas_dataframe": pd.DataFrame(
                np.random.randn(6, 4),
                index=pd.date_range("20130101", periods=6),
                columns=list("ABCD"),
            ),
        }
        return (ret,)


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="A message publisher and listener for native Python objects, NumPy arrays, and pandas series/dataframes."
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="publish",
        choices={"publish", "listen"},
        help="The transmission mode",
    )
    parser.add_argument(
        "--mware",
        type=str,
        default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission",
    )
    return parser.parse_args()


def main(args):
    """
    Main function to initiate Notifier class and communication.
    """
    notifier = Notifier()
    notifier.activate_communication(Notifier.exchange_object, mode=args.mode)

    while True:
        (msg_object,) = notifier.exchange_object(mware=args.mware)
        print("Method result:", msg_object)


if __name__ == "__main__":
    args = parse_args()
    main(args)



import os
import re


def get_project_info_from_setup():
    # when Wrapyfi is not installed
    try:
        curr_dir = os.path.dirname(__file__)
        setup_path = os.path.join(curr_dir, "..", "setup.py")
        with open(setup_path, "r") as f:
            content = f.read()
    except FileNotFoundError:
        return {}
    name_match = re.search(r"name\s*=\s*['\"]([^'\"]*)['\"]", content)
    version_match = re.search(r"version\s*=\s*['\"]([^'\"]*)['\"]", content)
    url_match = re.search(r"url\s*=\s*['\"]([^'\"]*)['\"]", content)
    doc_match = re.search(r"'Documentation':\s*['\"]([^'\"]*)['\"]", content)

    if not name_match or not version_match or not url_match:
        # raise RuntimeError("Unable to find name, version, or url string.")
        return {}

    return {
        "name": name_match.group(1),
        "version": version_match.group(1),
        "url": url_match.group(1),
        "doc": None if not doc_match else doc_match.group(1),
    }


# extract project info
project_info = get_project_info_from_setup()

__version__ = project_info.get("version", None)
__url__ = project_info.get("url", None)
__doc__ = project_info.get("doc", None)
name = project_info.get("name", "wrapyfi")

if __version__ is None or __url__ is None or __doc__ is None:
    try:
        from importlib import metadata

        mdata = metadata.metadata(__name__)
        __version__ = metadata.version(__name__)
        __url__ = mdata["Home-page"]
        # when installed with PyPi
        if __url__ is None:
            for url_extract in mdata.get_all("Project-URL"):
                __url__ = (
                    url_extract.split(", ")[1]
                    if url_extract.split(", ")[0] == "Homepage"
                    else __url__
                )
        if __doc__ is None:
            for url_extract in mdata.get_all("Project-URL"):
                __doc__ = (
                    url_extract.split(", ")[1]
                    if url_extract.split(", ")[0] == "Documentation"
                    else __doc__
                )
    except ImportError:
        try:
            # when Python < 3.8 and setuptools/pip have not been updated
            import pkg_resources

            mdata = pkg_resources.get_distribution(__name__).metadata
            __version__ = pkg_resources.require(__name__)[0].version
            __url__ = mdata["Home-page"]
            # when installed with PyPi
            if __url__ is None:
                for url_extract in mdata.get_all("Project-URL"):
                    __url__ = (
                        url_extract.split(", ")[1]
                        if url_extract.split(", ")[0] == "Homepage"
                        else __url__
                    )
            if __doc__ is None:
                for url_extract in mdata.get_all("Project-URL"):
                    __doc__ = (
                        url_extract.split(", ")[1]
                        if url_extract.split(", ")[0] == "Documentation"
                        else __doc__
                    )
        except pkg_resources.DistributionNotFound:
            __version__ = "unknown_version"
            __url__ = "unknown_url"
            __doc__ = "unknown_url"
    except Exception:
        __version__ = "unknown_version"
        __url__ = "unknown_url"
        __doc__ = "unknown_url"

from wrapyfi.utils import PluginRegistrar

PluginRegistrar.scan()

import logging

logging.getLogger().setLevel(logging.INFO)



import logging
import json
import time
import os
from typing import Optional, Tuple, Tuple

import numpy as np
import cv2
import zmq

from wrapyfi.connect.servers import Server, Servers
from wrapyfi.middlewares.zeromq import ZeroMQMiddlewareReqRep
from wrapyfi.encoders import JsonEncoder, JsonDecodeHook


SOCKET_IP = os.environ.get("WRAPYFI_ZEROMQ_SOCKET_IP", "127.0.0.1")
SOCKET_PUB_PORT = int(os.environ.get("WRAPYFI_ZEROMQ_SOCKET_REQ_PORT", 5558))
SOCKET_SUB_PORT = int(os.environ.get("WRAPYFI_ZEROMQ_SOCKET_REP_PORT", 5559))
START_PROXY_BROKER = (
    os.environ.get("WRAPYFI_ZEROMQ_START_PROXY_BROKER", True) != "False"
)
PROXY_BROKER_SPAWN = os.environ.get("WRAPYFI_ZEROMQ_PROXY_BROKER_SPAWN", "process")
WATCHDOG_POLL_REPEAT = None


if os.name == "nt" and PROXY_BROKER_SPAWN == "process" and START_PROXY_BROKER:
    PROXY_BROKER_SPAWN = "thread"
    logging.warning(
        "[ZeroMQ] Wrapyfi does not support multiprocessing on Windows. Please set "
        "the environment variable WRAPYFI_ZEROMQ_PROXY_BROKER_SPAWN='thread'. "
        "Switching automatically to 'thread' mode. "
    )


class ZeroMQServer(Server):
    def __init__(
        self,
        name: str,
        out_topic: str,
        carrier: str = "tcp",
        socket_ip: str = SOCKET_IP,
        socket_rep_port: int = SOCKET_PUB_PORT,
        socket_req_port: int = SOCKET_SUB_PORT,
        start_proxy_broker: bool = START_PROXY_BROKER,
        proxy_broker_spawn: bool = PROXY_BROKER_SPAWN,
        zeromq_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the server and start the device broker if necessary.

        :param name: str: Name of the server
        :param out_topic: str: Topics are not supported for the REQ/REP pattern in ZeroMQ. Any given topic is ignored
        :param carrier: str: Carrier protocol. ZeroMQ currently only supports TCP for REQ/REP pattern. Default is 'tcp'
        :param socket_ip: str: IP address of the socket. Default is '127.0.0.1'
        :param socket_rep_port: int: Port of the socket for REP pattern. Default is 5558
        :param socket_req_port: int: Port of the socket for REQ pattern. Default is 5559
        :param start_proxy_broker: bool: Whether to start a device broker. Default is True
        :param proxy_broker_spawn: str: Whether to spawn the device broker as a process or thread. Default is 'process'
        :param zeromq_kwargs: dict: Additional kwargs for the ZeroMQ Req/Rep middleware
        :param kwargs: dict: Additional kwargs for the server
        """
        if out_topic != "":
            logging.warning(
                f"[ZeroMQ] ZeroMQ does not support topics for the REQ/REP pattern. Topic {out_topic} removed"
            )
            out_topic = ""
        if carrier or carrier != "tcp":
            logging.warning(
                "[ZeroMQ] ZeroMQ does not support other carriers than TCP for REQ/REP pattern. Using TCP."
            )
            carrier = "tcp"
        super().__init__(name, out_topic, carrier=carrier, **kwargs)

        self.socket_rep_address = f"{carrier}://{socket_ip}:{socket_rep_port}"
        self.socket_req_address = f"{carrier}://{socket_ip}:{socket_req_port}"
        if start_proxy_broker:
            ZeroMQMiddlewareReqRep.activate(
                socket_rep_address=self.socket_rep_address,
                socket_req_address=self.socket_req_address,
                proxy_broker_spawn=proxy_broker_spawn,
                **zeromq_kwargs or {},
            )
        else:
            ZeroMQMiddlewareReqRep.activate(**zeromq_kwargs or {})

    def close(self):
        """
        Close the server.
        """
        if hasattr(self, "_socket") and self._socket:
            if self._socket is not None:
                self._socket.close()

    def __del__(self):
        self.close()


@Servers.register("NativeObject", "zeromq")
class ZeroMQNativeObjectServer(ZeroMQServer):
    def __init__(
        self,
        name: str,
        out_topic: str,
        carrier: str = "tcp",
        serializer_kwargs: Optional[dict] = None,
        deserializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Specific server handling native Python objects, serializing them to JSON strings for transmission.

        :param name: str: Name of the server
        :param out_topic: str: Topics are not supported for the REQ/REP pattern in ZeroMQ. Any given topic is ignored
        :param carrier: str: Carrier protocol. ZeroMQ currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, out_topic, carrier=carrier, **kwargs)
        self._plugin_encoder = JsonEncoder
        self._serializer_kwargs = serializer_kwargs or {}
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self.establish()

    def establish(self, **kwargs):
        """
        Establish the connection to the server.
        """
        self._socket = zmq.Context().instance().socket(zmq.REP)
        for socket_property in ZeroMQMiddlewareReqRep().zeromq_kwargs.items():
            if isinstance(socket_property[1], str):
                self._socket.setsockopt_string(
                    getattr(zmq, socket_property[0]), socket_property[1]
                )
            else:
                self._socket.setsockopt(
                    getattr(zmq, socket_property[0]), socket_property[1]
                )
        self._socket.connect(self.socket_req_address)
        self.established = True

    def await_request(self, *args, **kwargs):
        """
        Await and deserialize the client's request, returning the extracted arguments and keyword arguments.
        The method blocks until a message is received, then attempts to deserialize it using the configured JSON decoder
        hook, returning the extracted arguments and keyword arguments.

        :return: Tuple[list, dict]: A tuple containing two items:
                 - A list of arguments extracted from the received message
                 - A dictionary of keyword arguments extracted from the received message
        """
        message = self._socket.recv_string()
        try:
            request = json.loads(
                message,
                object_hook=self._plugin_decoder_hook,
                **self._deserializer_kwargs,
            )
            args, kwargs = request
            return args, kwargs
        except json.JSONDecodeError as e:
            logging.error(f"[ZeroMQ] Failed to decode message: {e}")
            return [], {}

    def reply(self, obj):
        """
        Serialize the provided Python object to a JSON string and send it as a reply to the client.
        The method uses the configured JSON encoder for serialization before sending the resultant string to the client.

        :param obj: Any: The Python object to be serialized and sent
        """
        obj_str = json.dumps(obj, cls=self._plugin_encoder, **self._serializer_kwargs)
        self._socket.send_string(obj_str)


@Servers.register("Image", "zeromq")
class ZeroMQImageServer(ZeroMQNativeObjectServer):

    def __init__(
        self,
        name: str,
        out_topic: str,
        carrier: str = "tcp",
        width: int = -1,
        height: int = -1,
        rgb: bool = True,
        fp: bool = False,
        jpg: bool = False,
        deserializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Specific server handling image data as numpy arrays, serializing them to JSON strings for transmission.

        :param name: str: Name of the server
        :param out_topic: str: Topics are not supported for the REQ/REP pattern in ZeroMQ. Any given topic is ignored
        :param carrier: str: Carrier protocol. ZeroMQ currently only supports TCP for REQ/REP pattern. Default is 'tcp'
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be decompressed from JPG. Default is False
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(
            name,
            out_topic,
            carrier=carrier,
            deserializer_kwargs=deserializer_kwargs,
            **kwargs,
        )

        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        self._type = np.float32 if self.fp else np.uint8

    def reply(self, img: np.ndarray):
        """
        Serialize the provided image data and send it as a reply to the client.

        :param img: np.ndarray: Image to send formatted as a cv2 image - np.ndarray[img_height, img_width, channels]
        """
        if img is None:
            logging.warning("[ZeroMQ] Image is None. Skipping reply.")
            return

        if (
            0 < self.width != img.shape[1]
            or 0 < self.height != img.shape[0]
            or not (
                (img.ndim == 2 and not self.rgb)
                or (img.ndim == 3 and self.rgb and img.shape[2] == 3)
            )
        ):
            raise ValueError("Incorrect image shape for server reply")

        if not img.flags["C_CONTIGUOUS"]:
            img = np.ascontiguousarray(img)

        if self.jpg:
            _, img_encoded = cv2.imencode(".jpg", img)
            img_bytes = img_encoded.tobytes()
            self._socket.send(img_bytes)
        else:
            img_list = img.tolist()
            img_json = json.dumps({"img": img_list})
            self._socket.send_string(img_json)


@Servers.register("AudioChunk", "zeromq")
class ZeroMQAudioChunkServer(ZeroMQNativeObjectServer):

    def __init__(
        self,
        name: str,
        out_topic: str,
        carrier: str = "tcp",
        channels: int = 1,
        rate: int = 44100,
        chunk: int = -1,
        deserializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Specific server handling audio data as numpy arrays, serializing them to JSON strings for transmission.

        :param name: str: Name of the server
        :param out_topic: str: Topics are not supported for the REQ/REP pattern in ZeroMQ. Any given topic is ignored
        :param carrier: str: Carrier protocol. ZeroMQ currently only supports TCP for REQ/REP pattern. Default is 'tcp'
        :param channels: int: Number of channels in the audio. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: Number of samples in the audio chunk. Default is -1 (use the chunk size of the received audio)
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(
            name,
            out_topic,
            carrier=carrier,
            deserializer_kwargs=deserializer_kwargs,
            **kwargs,
        )
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

    def reply(self, aud: Tuple[np.ndarray, int]):
        """
        Serialize the provided audio data and send it as a reply to the client.

        :param aud: Tuple[np.ndarray, int]: Audio chunk to publish formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        aud, rate = aud
        if 0 < self.rate != rate:
            raise ValueError("Incorrect audio rate for publisher")
        chunk, channels = aud.shape if len(aud.shape) > 1 else (aud.shape[0], 1)
        self.chunk = chunk if self.chunk == -1 else self.chunk
        self.channels = channels if self.channels == -1 else self.channels
        if 0 < self.chunk != chunk or 0 < self.channels != channels:
            raise ValueError("Incorrect audio shape for publisher")
        aud = np.require(aud, dtype=np.float32, requirements="C")

        aud_list = aud.tolist()
        aud_json = json.dumps({"aud": (int(chunk), int(channels), int(rate), aud_list)})
        self._socket.send_string(aud_json)



"""
Encoder and Decoder for Trax Array Data via Wrapyfi.

This script provides mechanisms to encode and decode Trax Array Data using Wrapyfi.
It utilizes base64 encoding and pickle serialization.

The script contains a class, `TraxArray`, registered as a plugin to manage the
conversion of Trax Array data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - Trax: A deep learning library that focuses on clear code and speed (refer to https://trax-ml.readthedocs.io/en/latest/notebooks/trax_intro.html for installation instructions)
        Note: If Trax is not available, HAVE_TRAX will be set to False and
        the plugin will be registered with no types. Trax uses JAX or TensorFlow-NumPy as its backend,
        so they must be installed as well. Trax installs JAX as a dependency, but TensorFlow must be installed separately.

    You can install the necessary packages using pip:
        ``pip install trax``  # Basic installation of Trax
"""

import pickle
import base64

from wrapyfi.utils import *

try:
    import trax
    import jax
    import jaxlib.xla_extension

    HAVE_TRAX = True
except ImportError:
    HAVE_TRAX = False


@PluginRegistrar.register(
    types=None if not HAVE_TRAX else jaxlib.xla_extension.ArrayImpl.__mro__[:-1]
)
class TraxArray(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the TraxArray plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode Trax Array data using pickle and base64.

        :param obj: jaxlib.xla_extension.ArrayImpl: The Trax Array data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, pickled data string, and any buffer data
        """
        buffers = []
        obj_data = pickle.dumps(obj, protocol=5, buffer_callback=buffers.append)
        obj_buffers = list(
            map(lambda x: base64.b64encode(memoryview(x)).decode("ascii"), buffers)
        )
        return True, dict(
            __wrapyfi__=(
                str(self.__class__.__name__),
                obj_data.decode("latin1"),
                *obj_buffers,
            )
        )

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a pickled and base64 encoded string back into Trax Array data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the pickled data string and any buffer data
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, pa.StructArray]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - jaxlib.xla_extension.ArrayImpl: The decoded Trax Array data
        """
        obj_data = obj_full[1].encode("latin1")
        obj_buffers = list(
            map(lambda x: base64.b64decode(x.encode("ascii")), obj_full[2:])
        )
        obj_data = bytearray(obj_data)
        for buf in obj_buffers:
            obj_data += buf
        return True, pickle.loads(obj_data, buffers=obj_buffers)



"""
iCub Head Controller and Camera Viewer using Wrapyfi for Communication

This script demonstrates the capability to control the iCub robot's head and view its camera feed using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern,
allowing for message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using Image messages for camera feed transmission.
    - Running publishers and listeners concurrently with the yarp.RFModule.
    - Using Wrapyfi for creating a port listener only.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - iCub Robot and Simulator: Ensure the robot and its simulator are installed and configured.
        When running in simulation mode, the `iCub_SIM` must be running in a standalone terminal
        (refer to the Wrapyfi documentation for installation instructions)
    - NumPy: Used for creating arrays (installed with Wrapyfi)
    - SciPy: For applying smoothing filters to the facial expressions (refer to https://www.scipy.org/install.html for installation instructions)
    - Pexpect: To control the facial expressions using RPC

    Install using pip:
    ``pip install "scipy==1.9.0" pexpect``

Run:
    # For the list of keyboard controls, refer to the comments in Keyboard Controls.

    # Alternative 1: Simulation Mode
        # Ensure that the `iCub_SIM` is running in a standalone terminal.

        # The listener displays images, and coordinates are published without utilizing Wrapyfi's utilities.

        ``python3 icub_head.py --simulation --get_cam_feed --control_head --control_expressions``

    # Alternative 2: Physical Robot
        # The listener displays images, and coordinates are published without utilizing Wrapyfi's utilities.

        ``python3 icub_head.py --get_cam_feed --control_head --control_expressions``

Keyboard Controls:
    - Head Control:
        - Up/Down: Control the head pitch
        - Right/Left: Control the head yaw
        - A/D: Control the head roll (right/left)
        - R: Reset the head to the initial position
        - Esc: Quit the application
    - Eye Control:
        - W/S: Control the eye pitch (up/down)
        - C/Z: Control the eye yaw (right/left)
        - R: Reset the eye to the initial position
        - Esc: Quit the application
    - Facial Expressions:
        - 0: Neutral
        - 1: Happy
        - 2: Sad
        - 3: Surprise
        - 4: Fear
        - 5: Disgust
        - 6: Anger
        - 7: Contempt
        - 8: Cunning
        - 9: Shy
        - Esc: Quit the application
    - Camera Feed:
        - Esc: Quit the application
"""

import os
import time
import argparse
import logging
from collections import deque

import cv2
import numpy as np
import yarp

try:
    import pexpect

    HAVE_PEXPECT = True
except ImportError:
    HAVE_PEXPECT = False

from wrapyfi.connect.wrapper import MiddlewareCommunicator

ICUB_DEFAULT_COMMUNICATOR = os.environ.get("WRAPYFI_DEFAULT_COMMUNICATOR", "yarp")
ICUB_DEFAULT_COMMUNICATOR = os.environ.get(
    "WRAPYFI_DEFAULT_MWARE", ICUB_DEFAULT_COMMUNICATOR
)
ICUB_DEFAULT_COMMUNICATOR = os.environ.get(
    "ICUB_DEFAULT_COMMUNICATOR", ICUB_DEFAULT_COMMUNICATOR
)
ICUB_DEFAULT_COMMUNICATOR = os.environ.get(
    "ICUB_DEFAULT_MWARE", ICUB_DEFAULT_COMMUNICATOR
)

EMOTION_LOOKUP = {
    "Neutral": [("LIGHTS", "neu")],
    "Happy": [("all", "hap")],
    "Sad": [("LIGHTS", "sad")],
    "Surprise": [("LIGHTS", "sur")],
    "Fear": [("raw", "L04"), ("raw", "R04"), ("raw", "M66")],  # change to array
    "Disgust": [("raw", "L01"), ("raw", "R01"), ("raw", "M66")],  # change to array
    "Anger": [("LIGHTS", "ang")],
    "Contempt": [("raw", "L01"), ("raw", "R09"), ("raw", "ME9")],  # change to array
    "Cunning": [("LIGHTS", "cun")],
    "Shy": [("LIGHTS", "shy")],
    "Evil": [("LIGHTS", "evi")],
}


def cartesian_to_spherical(xyz=None, x=None, y=None, z=None, expand_return=None):
    """
    Convert cartesian coordinates to spherical coordinates.

    :param xyz: tuple: Cartesian coordinates (x, y, z)
    :param x: float: Cartesian coordinate x
    :param y: float: Cartesian coordinate y
    :param z: float: Cartesian coordinate z
    :param expand_return: bool: Whether to return the spherical coordinates as a dictionary or not
    :return: tuple: Spherical coordinates (p, t, r) or dictionary: Spherical coordinates {"p": p, "t": t, "r": r}
    """
    from operator import xor
    import numpy as np

    assert xor((xyz is not None), all((x is not None, y is not None, z is not None)))

    if expand_return is None:
        expand_return = False if xyz is None else True
    if xyz is None:
        xyz = (x, y, z)

    ptr = np.zeros((3,))
    xy = xyz[0] ** 2 + xyz[1] ** 2
    ptr[0] = np.arctan2(xyz[1], xyz[0])
    ptr[1] = np.arctan2(
        xyz[2], np.sqrt(xy)
    )  # for elevation angle defined from XY-plane up
    # ptr[1] = np.arctan2(np.sqrt(xy), xyz[2])  # for elevation angle defined from Z-axis down
    ptr[2] = np.sqrt(xy + xyz[2] ** 2)
    return ptr if not expand_return else {"p": ptr[0], "t": ptr[1], "r": ptr[2]}


def mode_smoothing_filter(time_series, default, window_length=6, min_count=None):
    """
    Apply a smoothing filter to the time series using the mode of the last `window_length` values.

    :param time_series: list: Time series to apply the smoothing filter to
    :param default: object: Default value to return if the mode count is less than `min_count`
    :param window_length: int: Length of the window to apply the smoothing filter to
    :param min_count: int: Minimum number of values in the window to apply the smoothing filter
    """
    import scipy.stats

    if min_count is None:
        min_count = window_length // 2
    mode = scipy.stats.mode(time_series[-window_length:])
    return mode.mode[0] if mode.count >= min_count else default


class ICub(MiddlewareCommunicator, yarp.RFModule):
    """
    ICub head controller, facial expression transmitter and camera viewer. Head control can be achieved following two methods:
    1. Using the `control_head_gaze` method, which controls the head gaze in the spherical coordinate system.
    2. Using the `control_gaze_at_plane` method, which controls the head gaze in the cartesian coordinate system.
    Emotions can be controlled using the `update_facial_expressions` method.
    Camera feed can be viewed using the `receive_images` method.
    """

    MWARE = ICUB_DEFAULT_COMMUNICATOR
    CAP_PROP_FRAME_WIDTH = 320
    CAP_PROP_FRAME_HEIGHT = 240
    HEAD_COORDINATES_PORT = "/control_interface/head_coordinates"
    EYE_COORDINATES_PORT = "/control_interface/eye_coordinates"
    GAZE_PLANE_COORDINATES_PORT = "/control_interface/gaze_plane_coordinates"
    FACIAL_EXPRESSIONS_PORT = "/control_interface/facial_expressions"
    # constants
    FACIAL_EXPRESSIONS_QUEUE_SIZE = 7
    FACIAL_EXPRESSION_SMOOTHING_WINDOW = 6

    def __init__(
        self,
        simulation=False,
        headless=False,
        get_cam_feed=True,
        img_width=CAP_PROP_FRAME_WIDTH,
        img_height=CAP_PROP_FRAME_HEIGHT,
        control_head=True,
        set_head_coordinates=True,
        head_coordinates_port=HEAD_COORDINATES_PORT,
        control_eyes=True,
        set_eye_coordinates=True,
        eye_coordinates_port=EYE_COORDINATES_PORT,
        ikingaze=False,
        gaze_plane_coordinates_port=GAZE_PLANE_COORDINATES_PORT,
        control_expressions=False,
        set_facial_expressions=True,
        facial_expressions_port=FACIAL_EXPRESSIONS_PORT,
        mware=MWARE,
    ):
        """
        Initialize the ICub head controller, facial expression transmitter and camera viewer.

        :param simulation: bool: Whether to run the simulation or not
        :param headless: bool: Whether to run the headless mode or not
        :param get_cam_feed: bool: Whether to get (listen) the camera feed or not
        :param img_width: int: Width of the image
        :param img_height: int: Height of the image
        :param control_head: bool: Whether to control the head
        :param set_head_coordinates: bool: Whether to set (publish) the head coordinates
        :param head_coordinates_port: str: Port to receive the head coordinates for controlling the head
        :param control_eyes: bool: Whether to control the eyes
        :param set_eye_coordinates: bool: Whether to set (publish) the eye coordinates
        :param eye_coordinates_port: str: Port to receive the eye coordinates for controlling the eyes
        :param ikingaze: bool: Whether to use the iKinGazeCtrl
        :param gaze_plane_coordinates_port: str: Port to receive the gaze plane coordinates for controlling the head/eyes
        :param control_expressions: bool: Whether to control the facial expressions
        :param set_facial_expressions: bool: Whether to set (publish) the facial expressions
        :param facial_expressions_port: str: Port to receive the facial expressions for controlling the facial expressions
        :param mware: str: Middleware to use
        """

        self.__name__ = "iCubController"
        MiddlewareCommunicator.__init__(self)
        yarp.RFModule.__init__(self)

        self.MWARE = mware
        self.FACIAL_EXPRESSIONS_PORT = facial_expressions_port
        self.GAZE_PLANE_COORDINATES_PORT = gaze_plane_coordinates_port
        self.HEAD_COORDINATES_PORT = head_coordinates_port
        self.EYE_COORDINATES_PORT = eye_coordinates_port

        self.headless = headless
        self.ikingaze = ikingaze

        # prepare a property object
        props = yarp.Property()
        props.put("device", "remote_controlboard")
        props.put("local", "/client/head")

        if simulation:
            props.put("remote", "/icubSim/head")
            self.cam_props = {
                "cam_world_port": "/icubSim/cam",
                "cam_left_port": "/icubSim/cam/left",
                "cam_right_port": "/icubSim/cam/right",
            }
            emotion_cmd = f"yarp rpc /icubSim/face/emotions/in"
        else:
            props.put("remote", "/icub/head")
            self.cam_props = {
                "cam_world_port": "/icub/cam/left",
                "cam_left_port": "/icub/cam/left",
                "cam_right_port": "/icub/cam/right",
            }
            emotion_cmd = f"yarp rpc /icub/face/emotions/in"

        if img_width is not None:
            self.img_width = img_width
            self.CAP_PROP_FRAME_WIDTH = img_width
            self.cam_props["img_width"] = img_width

        if img_height is not None:
            self.img_height = img_height
            self.CAP_PROP_FRAME_HEIGHT = img_height
            self.cam_props["img_height"] = img_height

        if control_expressions:
            if HAVE_PEXPECT:
                # control emotional expressions using RPC
                self.client = pexpect.spawn(emotion_cmd)
            else:
                logging.error(
                    "pexpect must be installed to control the emotion interface"
                )
                self.activate_communication(self.update_facial_expressions, "disable")

            self.last_expression = [
                "",
                "",
            ]  # (emotion part on the robot's face , emotional expression category)
            self.expressions_queue = deque(maxlen=self.FACIAL_EXPRESSIONS_QUEUE_SIZE)
        else:
            self.activate_communication(self.update_facial_expressions, "disable")

        self._curr_eyes = [0, 0, 0]
        self._curr_head = [0, 0, 0]

        if control_head or control_eyes:
            if ikingaze:
                self._gaze_encs = yarp.Vector(3, 0.0)
                props_gaze = yarp.Property()
                props_gaze.clear()
                props_gaze.put("device", "gazecontrollerclient")
                props_gaze.put("remote", "/iKinGazeCtrl")
                props_gaze.put("local", "/client/gaze")
                #
                self._gaze_driver = yarp.PolyDriver(props_gaze)

                self._igaze = self._gaze_driver.viewIGazeControl()
                self._igaze.setStabilizationMode(True)

                # set movement speed
                # self.update_head_gaze_speed(head=0.8)
                # self.update_eye_gaze_speed(eye=0.5)
                self.activate_communication(self.control_head_gaze, "disable")
                self.activate_communication(self.control_eye_gaze, "disable")
                self.activate_communication(self._control_head_eye_gaze, "disable")

            else:
                # create remote driver
                self._head_driver = yarp.PolyDriver(props)

                # query motor control interfaces
                self._ipos = self._head_driver.viewIPositionControl()
                self._ienc = self._head_driver.viewIEncoders()

                # retrieve number of joints
                self._num_jnts = self._ipos.getAxes()

                logging.info(f"controlling {self._num_jnts} joints")

                # read encoders
                self._encs = yarp.Vector(self._num_jnts)
                self._ienc.getEncoders(self._encs.data())

                if not control_head:
                    self.activate_communication(self.control_head_gaze, "disable")
                    self.activate_communication(self.update_head_gaze_speed, "disable")
                if not control_eyes:
                    self.activate_communication(self.control_eye_gaze, "disable")
                    self.activate_communication(self.update_eye_gaze_speed, "disable")

                self.init_pos_head = yarp.Vector(self._num_jnts, self._encs.data())
                self.init_pos_eyes = yarp.Vector(self._num_jnts, self._encs.data())
                self.init_pos = yarp.Vector(self._num_jnts, self._encs.data())

                # set movement speed
                # self.update_head_gaze_speed(pitch=10.0, roll=10.0, yaw=20.0)
                # self.update_eye_gaze_speed(pitch=10.0, yaw=10.0, vergence=20.0)

        else:
            self.activate_communication(self.reset_gaze, "disable")
            self.activate_communication(self.update_head_gaze_speed, "disable")
            self.activate_communication(self.control_head_gaze, "disable")
            self.activate_communication(self.update_eye_gaze_speed, "disable")
            self.activate_communication(self.control_eye_gaze, "disable")
            self.activate_communication(self._control_head_eye_gaze, "disable")
            self.activate_communication(self.wait_for_gaze, "disable")
            self.activate_communication(self.control_gaze_at_plane, "disable")

        if get_cam_feed:
            # control the listening properties from within the app
            self.activate_communication(self.receive_images, "listen")
        if facial_expressions_port:
            if set_facial_expressions:
                self.activate_communication(self.acquire_facial_expressions, "publish")
            else:
                for _ in range(self.FACIAL_EXPRESSIONS_QUEUE_SIZE):
                    self.update_facial_expressions("hap", part="all", smoothing="mode")
                self.activate_communication(self.acquire_facial_expressions, "listen")
        if head_coordinates_port:
            if set_head_coordinates:
                self.activate_communication(self.acquire_head_coordinates, "publish")
            else:
                self.activate_communication(self.acquire_head_coordinates, "listen")
        if eye_coordinates_port:
            if set_eye_coordinates:
                self.activate_communication(self.acquire_eye_coordinates, "publish")
            else:
                self.activate_communication(self.acquire_eye_coordinates, "listen")
        if gaze_plane_coordinates_port:
            self.activate_communication(self.control_gaze_at_plane, "listen")

        self.build()

    def build(self):
        """
        Updates the default method arguments according to constructor arguments. This method is called by the module constructor.
        It is not necessary to call it manually.
        """
        ICub.acquire_head_coordinates.__defaults__ = (
            self.HEAD_COORDINATES_PORT,
            None,
            self.MWARE,
        )
        ICub.acquire_eye_coordinates.__defaults__ = (
            self.EYE_COORDINATES_PORT,
            None,
            self.MWARE,
        )
        ICub.receive_gaze_plane_coordinates.__defaults__ = (
            self.GAZE_PLANE_COORDINATES_PORT,
            self.MWARE,
        )
        ICub.wait_for_gaze.__defaults__ = (True, self.MWARE)
        ICub.reset_gaze.__defaults__ = (self.MWARE,)
        ICub.update_head_gaze_speed.__defaults__ = (10.0, 10.0, 20.0, 0.8, self.MWARE)
        ICub.control_head_gaze.__defaults__ = (0.0, 0.0, 0.0, "xyz", self.MWARE)
        ICub.update_eye_gaze_speed.__defaults__ = (10.0, 10.0, 20.0, 0.5, self.MWARE)
        ICub.control_eye_gaze.__defaults__ = (0.0, 0.0, 0.0, self.MWARE)
        ICub._control_head_eye_gaze.__defaults__ = (self.MWARE,)
        ICub.control_gaze_at_plane.__defaults__ = (
            0,
            0,
            0.3,
            0.3,
            True,
            True,
            self.MWARE,
        )
        ICub.acquire_facial_expressions.__defaults__ = (
            self.FACIAL_EXPRESSIONS_PORT,
            None,
            self.MWARE,
        )
        ICub.update_facial_expressions.__defaults__ = (None, False, "mode", self.MWARE)
        ICub.receive_images.__defaults__ = (
            self.CAP_PROP_FRAME_WIDTH,
            self.CAP_PROP_FRAME_HEIGHT,
            True,
        )

    @MiddlewareCommunicator.register(
        "NativeObject", "$_mware", "ICub", "$head_coordinates_port", should_wait=False
    )
    def acquire_head_coordinates(
        self,
        head_coordinates_port=HEAD_COORDINATES_PORT,
        cv2_key=None,
        _mware=MWARE,
        **kwargs,
    ):
        """
        Acquire head coordinates for controlling the iCub.

        :param head_coordinates_port: str: Port to receive head coordinates
        :param cv2_key: int: Key pressed by the user
        :return: dict: Head orientation coordinates
        """

        if cv2_key is None:
            logging.error("controlling orientation in headless mode not yet supported")
            return (None,)
        else:
            if cv2_key == 27:  # Esc key to exit
                exit(0)
            elif cv2_key == -1:  # normally -1 returned,so don't print it
                pass
            # the keyboard commands for controlling the robot
            elif cv2_key == 82:  # Up key
                self._curr_head[0] += 1
                logging.info("head pitch up")
            elif cv2_key == 84:  # Down key
                self._curr_head[0] -= 1
                logging.info("head pitch down")
            elif cv2_key == 83:  # Right key
                self._curr_head[2] -= 1
                logging.info("head yaw left")
            elif cv2_key == 81:  # Left key
                self._curr_head[2] += 1
                logging.info("head yaw right")
            elif cv2_key == 97:  # A key
                self._curr_head[1] -= 1
                logging.info("head roll right")
            elif cv2_key == 100:  # D key
                self._curr_head[1] += 1
                logging.info("head roll left")
            elif cv2_key == 114:  # R key: reset the orientation
                self._curr_head = [0, 0, 0]
                self.reset_gaze()
                logging.info("resetting the orientation")
            else:
                logging.info(cv2_key)  # else print its value
                return (None,)

            return (
                {
                    "topic": head_coordinates_port.split("/")[-1],
                    "timestamp": time.time(),
                    "pitch": self._curr_head[0],
                    "roll": self._curr_head[1],
                    "yaw": self._curr_head[2],
                    "order": "zyx",
                },
            )

    @MiddlewareCommunicator.register(
        "NativeObject", "$_mware", "ICub", "$eye_coordinates_port", should_wait=False
    )
    def acquire_eye_coordinates(
        self,
        eye_coordinates_port=EYE_COORDINATES_PORT,
        cv2_key=None,
        _mware=MWARE,
        **kwargs,
    ):
        """
        Acquire eye coordinates for controlling the iCub.

        :param eye_coordinates_port: str: Port to receive eye coordinates
        :param cv2_key: int: Key pressed by the user
        :return: dict: Eye oreintation coordinates
        """

        if cv2_key is None:
            logging.error("controlling orientation in headless mode not yet supported")
            return (None,)
        else:
            if cv2_key == 27:  # Esc key to exit
                exit(0)
            elif cv2_key == -1:  # normally -1 returned,so don't print it
                pass
            # the keyboard commands for controlling the robot
            elif cv2_key == 119:  # W key
                self._curr_eyes[0] += 1
                logging.info("eye pitch up")
            elif cv2_key == 115:  # S key
                self._curr_eyes[0] -= 1
                logging.info("eye pitch down")
            elif cv2_key == 122:  # Z key
                self._curr_eyes[1] -= 1
                logging.info("eye yaw left")
            elif cv2_key == 99:  # C key
                self._curr_eyes[1] += 1
                logging.info("eye yaw right")
            elif cv2_key == 114:  # R key: reset the orientation
                self._curr_eyes = [0, 0, 0]
                self.reset_gaze()
                logging.info("resetting the orientation")
            else:
                logging.info(cv2_key)  # else print its value
                return (None,)

            return (
                {
                    "topic": eye_coordinates_port.split("/")[-1],
                    "timestamp": time.time(),
                    "pitch": self._curr_eyes[0],
                    "yaw": self._curr_eyes[1],
                    "vergence": self._curr_eyes[2],
                },
            )

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ICub",
        "$gaze_plane_coordinates_port",
        should_wait=False,
    )
    def receive_gaze_plane_coordinates(
        self,
        gaze_plane_coordinates_port=GAZE_PLANE_COORDINATES_PORT,
        _mware=MWARE,
        **kwargs,
    ):
        """
        Receive gaze plane (normalized x,y) coordinates for controlling the iCub.

        :param gaze_plane_coordinates_port: str: Port to receive gaze plane coordinates
        :return: dict: Gaze plane coordinates
        """
        return (None,)

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ICub",
        "/icub_controller/logs/wait_for_gaze",
        should_wait=False,
    )
    def wait_for_gaze(self, reset=True, _mware=MWARE):
        """
        Wait for the gaze actuation to complete.
        :param reset: bool: Whether to reset the gaze location (centre)
        :param _mware: str: Middleware to use
        :return: dict: Gaze waiting log for a given time step
        """
        if self.ikingaze:
            # self._igaze.clearNeckPitch()
            # self._igaze.clearNeckRoll()
            # self._igaze.clearNeckYaw()
            # self._igaze.clearEyes()
            if reset:
                self._igaze.lookAtAbsAngles(self._gaze_encs)
            self._igaze.waitMotionDone(timeout=2.0)
        else:
            if reset:
                self._ipos.positionMove(self._encs.data())
                while not self._ipos.checkMotionDone():
                    pass
        return (
            {
                "topic": "logging_wait_for_gaze",
                "timestamp": time.time(),
                "command": f"waiting for gaze completed with reset={reset}",
            },
        )

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ICub",
        "/icub_controller/logs/reset_gaze",
        should_wait=False,
    )
    def reset_gaze(self, _mware=MWARE):
        """
        Reset the eyes and head to their original position.

        :param _mware: str: Middleware to use
        :return: dict: Gaze reset log for a given time step
        """
        self.wait_for_gaze(reset=True)
        return (
            {
                "topic": "logging_reset_gaze",
                "timestamp": time.time(),
                "command": f"reset gaze",
            },
        )

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "ICub",
        "/icub_controller/logs/head_speed",
        should_wait=False,
    )
    def update_head_gaze_speed(
        self, pitch=10.0, roll=10.0, yaw=20.0, head=0.8, _mware=MWARE, **kwargs
    ):
        """
        Control the iCub head speed.

        :param pitch: float->pitch[deg/s]: Pitch speed
        :param roll: float->roll[deg/s]: Roll speed
        :param yaw: float->yaw[deg/s]: Yaw speed
        :param head: float->speed[0,1]: Neck trajectory speed in normalized units (only when using iKinGazeCtrl)
        :param _mware: str: Middleware to use
        :return: dict: Head orientation speed log for a given time step
        """
        if self.ikingaze:
            self._igaze.setNeckTrajTime(head)

            return (
                {
                    "topic": "logging_head_speed",
                    "timestamp": time.time(),
                    "command": f"head speed set to {head}",
                },
            )
        else:
            self._ipos.setRefSpeed(0, pitch)
            self._ipos.setRefSpeed(1, roll)
            self._ipos.setRefSpeed(2, yaw)

            return (
                {
                    "topic": "logging_head_speed",
                    "timestamp": time.time(),
                    "command": f"head speed set to {pitch, roll, yaw} (pitch, roll, yaw)",
                },
            )

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "ICub",
        "/icub_controller/logs/eye_speed",
        should_wait=False,
    )
    def update_eye_gaze_speed(
        self, pitch=10.0, yaw=10.0, vergence=20.0, eye=0.5, _mware=MWARE, **kwargs
    ):
        """
        Control the iCub eye speed.

        :param pitch: float->pitch[deg/s]: Pitch speed
        :param yaw: float->yaw[deg/s]: Yaw speed
        :param vergence: float->vergence[deg/s]: Speed of vergence shift between the eyes
        :param eye: float->speed[0,1]: Eye trajectory speed in normalized units (only when using iKinGazeCtrl)
        :param _mware: str: Middleware to use
        :return: dict: Eye orientation speed log for a given time step
        """
        if self.ikingaze:
            self._igaze.setEyesTrajTime(eye)

            return (
                {
                    "topic": "logging_eye_speed",
                    "timestamp": time.time(),
                    "command": f"eye speed set to {eye}",
                },
            )
        else:
            self._ipos.setRefSpeed(3, pitch)
            self._ipos.setRefSpeed(4, yaw)
            self._ipos.setRefSpeed(5, vergence)

            return (
                {
                    "topic": "logging_eye_speed",
                    "timestamp": time.time(),
                    "command": f"eye speed set to {pitch, yaw, vergence} (pitch, yaw, vergence)",
                },
            )

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ICub",
        "/icub_controller/logs/head_orientation_coordinates",
        should_wait=False,
    )
    def control_head_gaze(
        self, pitch=0.0, roll=0.0, yaw=0.0, order="xyz", _mware=MWARE, **kwargs
    ):
        """
        Control the iCub head relative to previous coordinates following the roll,pitch,yaw convention (order=xyz)
        (initialized at 0 looking straight ahead).

        :param pitch: float->pitch[deg]: Pitch angle
        :param roll: float->roll[deg]: Roll angle
        :param yaw: float->yaw[deg]: Yaw angle
        :param order: str: Euler axis order. Only accepts xyz (roll, pitch, yaw)
        :param _mware: str: Middleware to use
        :return: dict: Head orientation coordinates log for a given time step
        """
        if order != "xyz":
            logging.error(
                "only accepts ratation angles following the order='xyz' convention"
            )
            return (None,)
        # wait for the action to complete
        # self.wait_for_gaze(reset=False)

        # initialize a new tmp vector identical to encs
        self.init_pos_head = yarp.Vector(self._num_jnts, self._encs.data())

        # head control
        self.init_pos_head.set(0, self.init_pos_head.get(0) + pitch)  # tilt/pitch
        self.init_pos_head.set(1, self.init_pos_head.get(1) + roll)  # swing/roll
        self.init_pos_head.set(2, self.init_pos_head.get(2) + yaw)  # pan/yaw

        # self._ipos.positionMove(self.init_pos_head.data())
        self._curr_head = list((pitch, roll, yaw))

        return (
            {
                "topic": "logging_head_coordinates",
                "timestamp": time.time(),
                "command": f"head orientation set to {self._curr_head} (pitch, roll, yaw)",
            },
        )

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ICub",
        "/icub_controller/logs/eye_orientation_coordinates",
        should_wait=False,
    )
    def control_eye_gaze(
        self, pitch=0.0, yaw=0.0, vergence=0.0, _mware=MWARE, **kwargs
    ):
        """
        Control the iCub eyes relative to previous coordinates (initialized at 0 looking straight ahead).

        :param pitch: float->pitch[deg]: Pitch angle
        :param yaw: float->yaw[deg]: Yaw (version) angle
        :param vergence: float->yaw[deg]: Vergence angle between the eyes
        :param _mware: str: Middleware to use
        :return: dict: Eye orientation coordinates log for a given time step
        """
        # wait for the action to complete
        # self.wait_for_gaze(reset=False)

        # initialize a new tmp vector identical to encs
        self.init_pos_eyes = yarp.Vector(self._num_jnts, self._encs.data())

        # eye control
        self.init_pos_eyes.set(3, self.init_pos_eyes.get(3) + pitch)  # eye tilt
        self.init_pos_eyes.set(4, self.init_pos_eyes.get(4) + yaw)  # eye pan/version
        self.init_pos_eyes.set(
            5, self.init_pos_eyes.get(5) + vergence
        )  # the vergence between the eyes (to align, set to 0)

        # self._ipos.positionMove(self.init_pos_eyes.data())
        self._curr_eyes = list((pitch, yaw, vergence))

        return (
            {
                "topic": "logging_eye_coordinates",
                "timestamp": time.time(),
                "command": f"eye orientation set to {self._curr_eyes} (pitch, yaw, vergence)",
            },
        )

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ICub",
        "/icub_controller/logs/head_eye_orientation_coordinates",
        should_wait=False,
    )
    def _control_head_eye_gaze(self, _mware=MWARE, **kwargs):
        """
        Issue the movement command

        :param _mware: str: Middleware to use
        :return: dict: Head orientation coordinates log for a given time step
        """

        # initialize a new tmp vector identical to encs
        self.init_pos = yarp.Vector(self._num_jnts, self._encs.data())

        # head + eye control
        self.init_pos.set(0, self.init_pos_head.get(0))  # tilt/pitch
        self.init_pos.set(1, self.init_pos_head.get(1))  # swing/roll
        self.init_pos.set(2, self.init_pos_head.get(2))  # pan/yaw
        self.init_pos.set(3, self.init_pos_eyes.get(3))  # eye tilt
        self.init_pos.set(4, self.init_pos_eyes.get(4))  # eye pan/version
        self.init_pos.set(
            5, self.init_pos_eyes.get(5)
        )  # the vergence between the eyes (to align, set to 0)
        self._ipos.positionMove(self.init_pos.data())

        return (
            {
                "topic": "logging_head_eye_coordinates",
                "timestamp": time.time(),
                "command": f"head orientation set to {self._curr_head} (pitch, roll, yaw) and eye orientation to {self._curr_eyes} (pitch, yaw, vergence)",
            },
        )

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ICub",
        "/icub_controller/logs/gaze_plane_coordinates",
        should_wait=False,
    )
    def control_gaze_at_plane(
        self,
        x=0.0,
        y=0.0,
        limit_x=0.3,
        limit_y=0.3,
        control_eyes=True,
        control_head=True,
        _mware=MWARE,
        **kwargs,
    ):
        """
        Gaze at specific point in a normalized plane in front of the iCub.

        :param x: float->x[-1,1]: x coordinate in the plane limited to the range of -1 (left) and 1 (right)
        :param y: float->y[-1,1]: y coordinate in the plane limited to the range of -1 (bottom) and 1 (top)
        :param limit_x: float->limit_x[0,1]: x coordinate limit in the plane
        :param limit_y: float->limit_y[0,1]: y coordinate limit in the plane
        :param control_eyes: bool: Whether to control the eyes of the robot directly
        :param control_head: bool: Whether to control the head of the robot directly
        :return: dict: Gaze coordinates log for a given time step
        """
        # wait for the action to complete
        # self.wait_for_gaze(reset=False)

        xy = np.array((x, y)) * np.array((limit_x, limit_y))  # limit viewing region
        ptr = cartesian_to_spherical(x=1, y=xy[0], z=-xy[1], expand_return=False)
        # initialize a new tmp vector identical to encs
        ptr_degrees = (np.rad2deg(ptr[0]), np.rad2deg(ptr[1]))

        if control_eyes and control_head:
            if not self.ikingaze:
                logging.error(
                    "set ikingaze=True in order to move eyes and head simultaneously"
                )
                return (None,)
            self.init_pos_ikin = yarp.Vector(3, self._gaze_encs.data())
            self.init_pos_ikin.set(0, ptr_degrees[0])
            self.init_pos_ikin.set(1, ptr_degrees[1])
            self.init_pos_ikin.set(2, 0.0)
            self._igaze.lookAtAbsAngles(self.init_pos_ikin)

        elif control_head:
            if self.ikingaze:
                logging.error("set ikingaze=False in order to move head only")
                return (None,)
            self.control_head_gaze(pitch=ptr_degrees[1], roll=0, yaw=ptr_degrees[0])
        elif control_eyes:
            if self.ikingaze:
                logging.error("set ikingaze=False in order to move eyes only")
                return (None,)
            self.control_eye_gaze(pitch=ptr_degrees[1], yaw=ptr_degrees[0], vergence=0)

        return (
            {
                "topic": "logging_gaze_plane_coordinates",
                "timestamp": time.time(),
                "command": f"moving gaze toward {ptr_degrees} with head={control_head} and eyes={control_eyes}",
            },
        )

    @MiddlewareCommunicator.register(
        "NativeObject", "$_mware", "ICub", "$facial_expressions_port", should_wait=False
    )
    def acquire_facial_expressions(
        self,
        facial_expressions_port=FACIAL_EXPRESSIONS_PORT,
        cv2_key=None,
        _mware=MWARE,
        **kwargs,
    ):
        """
        Acquire facial expressions from the iCub.

        :param facial_expressions_port: str: Port to acquire facial expressions from
        :param cv2_key: int: Key to press to set the facial expression
        :return: dict: Facial expressions log for a given time step
        """
        emotion = None
        if cv2_key is None:
            logging.error("controlling expressions in headless mode not yet supported")
            return (None,)
        else:
            if cv2_key == 27:  # Esc key to exit
                exit(0)
            elif cv2_key == -1:  # normally -1 returned,so don"t print it
                pass
            elif cv2_key == 48:  # 0 key: Neutral emotion
                emotion = "Neutral"
                logging.info("expressing neutrality")
            elif cv2_key == 49:  # 1 key: Happy emotion
                emotion = "Happy"
                logging.info("expressing happiness")
            elif cv2_key == 50:  # 2 key: Sad emotion
                emotion = "Sad"
                logging.info("expressing sadness")
            elif cv2_key == 51:  # 3 key: Surprise emotion
                emotion = "Surprise"
                logging.info("expressing surprise")
            elif cv2_key == 52:  # 4 key: Fear emotion
                emotion = "Fear"
                logging.info("expressing fear")
            elif cv2_key == 53:  # 5 key: Disgust emotion
                emotion = "Disgust"
                logging.info("expressing disgust")
            elif cv2_key == 54:  # 6 key: Anger emotion
                emotion = "Anger"
                logging.info("expressing anger")
            elif cv2_key == 55:  # 7 key: Contempt emotion
                emotion = "Contempt"
                logging.info("expressing contempt")
            elif cv2_key == 56:  # 8 key: Cunning emotion
                emotion = "Cunning"
                logging.info("expressing cunningness")
            elif cv2_key == 57:  # 9 key: Shy emotion
                emotion = "Shy"
                logging.info("expressing shyness")
            else:
                logging.info(cv2_key)  # else print its value
                return (None,)
            return (
                {
                    "topic": facial_expressions_port.split("/")[-1],
                    "timestamp": time.time(),
                    "emotion_category": emotion,
                },
            )

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ICub",
        "/icub_controller/logs/facial_expressions",
        should_wait=False,
    )
    def update_facial_expressions(
        self, expression, part=False, smoothing="mode", _mware=MWARE, **kwargs
    ):
        """
        Control facial expressions of the iCub.

        :param expression: str: Expression to be controlled
        :param expression: str or tuple(str->part, str->emotion) or list[str] or list[tuple(str->part, str->emotion)]:
                            Expression/s abbreviation or matching lookup table entry.
                            If a list is provided, the actions are executed in sequence
        :param part: str: Abbreviation describing parts to control (refer to iCub documentation) ( mou, eli, leb, reb, all, raw, LIGHTS)
        :param smoothing: str: Name of smoothing filter to avoid abrupt changes in emotional expressions
        :return: Emotion log for a given time step
        """
        if expression is None:
            return (None,)

        if isinstance(expression, (list, tuple)):
            expression = expression[-1]

        if smoothing == "mode":
            self.expressions_queue.append(expression)
            transmitted_expression = mode_smoothing_filter(
                list(self.expressions_queue),
                default="neu",
                window_length=self.FACIAL_EXPRESSION_SMOOTHING_WINDOW,
            )
        else:
            transmitted_expression = expression

        expressions_lookup = EMOTION_LOOKUP.get(
            transmitted_expression, transmitted_expression
        )
        if isinstance(expressions_lookup, str):
            expressions_lookup = [(part if part else "LIGHTS", expressions_lookup)]

        if (
            self.last_expression[0] == (part if part else "LIGHTS")
            and self.last_expression[1] == transmitted_expression
        ):
            expressions_lookup = []

        for part_lookup, expression_lookup in expressions_lookup:
            if part_lookup == "LIGHTS":
                self.client.sendline(f"set leb {expression_lookup}")
                self.client.expect(">>")
                self.client.sendline(f"set reb {expression_lookup}")
                self.client.expect(">>")
                self.client.sendline(f"set mou {expression_lookup}")
                self.client.expect(">>")
                logging.info(f"set leb/reb/mou {expression_lookup}")
            else:
                self.client.sendline(f"set {part_lookup} {expression_lookup}")
                self.client.expect(">>")
                logging.info(f"set {part_lookup} {expression_lookup}")

        self.last_expression[0] = part
        self.last_expression[1] = transmitted_expression

        return (
            {
                "topic": "logging_facial_expressions",
                "timestamp": time.time(),
                "command": f"emotion set to {part} {expression} with smoothing={smoothing}",
            },
        )

    @MiddlewareCommunicator.register(
        "Image",
        "yarp",
        "ICub",
        "$cam_world_port",
        width="$img_width",
        height="$img_height",
        rgb="$_rgb",
    )
    @MiddlewareCommunicator.register(
        "Image",
        "yarp",
        "ICub",
        "$cam_left_port",
        width="$img_width",
        height="$img_height",
        rgb="$_rgb",
    )
    @MiddlewareCommunicator.register(
        "Image",
        "yarp",
        "ICub",
        "$cam_right_port",
        width="$img_width",
        height="$img_height",
        rgb="$_rgb",
    )
    def receive_images(
        self,
        cam_world_port,
        cam_left_port,
        cam_right_port,
        img_width=CAP_PROP_FRAME_WIDTH,
        img_height=CAP_PROP_FRAME_HEIGHT,
        _rgb=True,
    ):
        """
        Receive images from the iCub.

        :param cam_world_port: str: Port to receive images from the world camera
        :param cam_left_port: str: Port to receive images from the left camera
        :param cam_right_port: str: Port to receive images from the right camera
        :param img_width: int: Width of the image
        :param img_height: int: Height of the image
        :param _rgb: bool: Whether the image is RGB or not
        :return: Images from the iCub
        """
        external_cam, left_cam, right_cam = None, None, None
        return external_cam, left_cam, right_cam

    def getPeriod(self):
        """
        Get the period of the module.
        :return: float: Period of the module
        """
        return 0.01

    def updateModule(self):
        # print(self.getPeriod())
        external_cam, left_cam, right_cam = self.receive_images(**self.cam_props)
        if external_cam is None:
            external_cam = np.zeros((self.img_height, self.img_width, 1), dtype="uint8")
            left_cam = np.zeros((self.img_height, self.img_width, 1), dtype="uint8")
            right_cam = np.zeros((self.img_height, self.img_width, 1), dtype="uint8")
        else:
            external_cam = cv2.cvtColor(external_cam, cv2.COLOR_BGR2RGB)
            left_cam = cv2.cvtColor(left_cam, cv2.COLOR_BGR2RGB)
            right_cam = cv2.cvtColor(right_cam, cv2.COLOR_BGR2RGB)
        if not self.headless:
            cv2.imshow(
                "ICubCam", np.concatenate((left_cam, external_cam, right_cam), axis=1)
            )
            k = cv2.waitKey(30)
        else:
            k = None

        (switch_emotion,) = self.acquire_facial_expressions(
            facial_expressions_port=self.FACIAL_EXPRESSIONS_PORT,
            cv2_key=k,
            _mware=self.MWARE,
        )
        if switch_emotion is not None and isinstance(switch_emotion, dict):
            self.update_facial_expressions(
                switch_emotion.get("emotion_category", None),
                part=switch_emotion.get("part", False),
                _mware=self.MWARE,
            )

        # move robot head
        (move_robot_head,) = self.acquire_head_coordinates(
            head_coordinates_port=self.HEAD_COORDINATES_PORT,
            cv2_key=k,
            _mware=self.MWARE,
        )
        if move_robot_head is not None and isinstance(move_robot_head, dict):
            robot_head_speed = move_robot_head.get("speed", False)
            if robot_head_speed and isinstance(robot_head_speed, dict):
                self.update_head_gaze_speed(
                    pitch=robot_head_speed.get("pitch", 10.0),
                    roll=robot_head_speed.get("roll", 10.0),
                    yaw=robot_head_speed.get("yaw", 20.0),
                    _mware=self.MWARE,
                )
            if move_robot_head.get("reset_gaze", False):
                self.reset_gaze()
            self.control_head_gaze(
                pitch=move_robot_head.get("pitch", 0.0),
                roll=move_robot_head.get("roll", 0.0),
                yaw=move_robot_head.get("yaw", 0.0),
                _mware=self.MWARE,
            )

        # move robot eyes
        (move_robot_eyes,) = self.acquire_eye_coordinates(
            eye_coordinates_port=self.EYE_COORDINATES_PORT, cv2_key=k, _mware=self.MWARE
        )
        if move_robot_eyes is not None and isinstance(move_robot_eyes, dict):
            robot_eye_speed = move_robot_eyes.get("speed", False)
            if robot_eye_speed and isinstance(robot_eye_speed, dict):
                self.update_eye_gaze_speed(
                    pitch=robot_eye_speed.get("pitch", 10.0),
                    yaw=robot_eye_speed.get("yaw", 10.0),
                    vergence=robot_eye_speed.get("vergence", 20.0),
                    _mware=self.MWARE,
                )
            if move_robot_eyes.get("reset_gaze", False):
                self.reset_gaze()
            self.control_eye_gaze(
                pitch=move_robot_eyes.get("pitch", 0.0),
                yaw=move_robot_eyes.get("yaw", 0.0),
                vergence=move_robot_eyes.get("vergence", 0.0),
                _mware=self.MWARE,
            )

        if move_robot_head is not None or move_robot_eyes is not None:
            self._control_head_eye_gaze()

        (move_robot,) = self.receive_gaze_plane_coordinates(
            gaze_plane_coordinates_port=self.GAZE_PLANE_COORDINATES_PORT,
            _mware=self.MWARE,
        )
        if move_robot is not None and isinstance(move_robot, dict):
            robot_eye_speed = move_robot.get("eye_speed", False)
            if robot_eye_speed and isinstance(robot_eye_speed, dict):
                self.update_eye_gaze_speed(
                    **(
                        {
                            "pitch": robot_eye_speed.get("pitch", 10.0),
                            "yaw": robot_eye_speed.get("yaw", 10.0),
                            "vergence": robot_eye_speed.get("vergence", 20.0),
                            "_mware": self.MWARE,
                        }
                        if not self.ikingaze
                        else {
                            "eye": robot_eye_speed.get("eye", 0.5),
                            "_mware": self.MWARE,
                        }
                    )
                )

            robot_head_speed = move_robot.get("head_speed", False)
            if robot_head_speed and isinstance(robot_head_speed, dict):
                self.update_head_gaze_speed(
                    **(
                        {
                            "pitch": robot_head_speed.get("pitch", 10.0),
                            "roll": robot_head_speed.get("roll", 10.0),
                            "yaw": robot_head_speed.get("yaw", 20.0),
                            "_mware": self.MWARE,
                        }
                        if not self.ikingaze
                        else {
                            "head": robot_head_speed.get("head", 0.8),
                            "_mware": self.MWARE,
                        }
                    )
                )

            if move_robot.get("reset_gaze", False):
                self.reset_gaze()
            self.control_gaze_at_plane(
                x=move_robot.get("x", 0.0),
                y=move_robot.get("y", 0.0),
                limit_x=move_robot.get("limit_x", 0.3),
                limit_y=move_robot.get("limit_y", 0.3),
                control_head=move_robot.get(
                    "control_head", False if not self.ikingaze else True
                ),
                control_eyes=move_robot.get("control_eyes", True),
                _mware=self.MWARE,
            ),

        return True


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--simulation", action="store_true", help="Run in simulation")
    parser.add_argument("--headless", action="store_true", help="Disable CV2 GUI")
    parser.add_argument("--ikingaze", action="store_true", help="Enable iKinGazeCtrl")
    parser.add_argument(
        "--get_cam_feed",
        action="store_true",
        help="Get the camera feeds from the robot",
    )
    parser.add_argument("--control_head", action="store_true", help="Control the head")
    parser.add_argument(
        "--set_head_coordinates",
        action="store_true",
        help="Publish head coordinates set using keyboard commands",
    )
    parser.add_argument(
        "--head_coordinates_port",
        type=str,
        default="",
        help="The port (topic) name used for receiving and transmitting head orientation "
        "Setting the port name without --set_head_coordinates will only receive the coordinates",
    )
    parser.add_argument("--control_eyes", action="store_true", help="Control the eyes")
    parser.add_argument(
        "--set_eye_coordinates",
        action="store_true",
        help="Publish eye coordinates set using keyboard commands",
    )
    parser.add_argument(
        "--eye_coordinates_port",
        type=str,
        default="",
        help="The port (topic) name used for receiving and transmitting eye orientation "
        "Setting the port name without --set_eye_coordinates will only receive the coordinates",
    )
    parser.add_argument(
        "--gaze_plane_coordinates_port",
        type=str,
        default="",
        help="The port (topic) name used for receiving plane coordinates in 2D for robot to look at",
    )
    parser.add_argument(
        "--control_expressions",
        action="store_true",
        help="Control the facial expressions",
    )
    parser.add_argument(
        "--set_facial_expressions",
        action="store_true",
        help="Publish facial expressions set using keyboard commands",
    )
    parser.add_argument(
        "--facial_expressions_port",
        type=str,
        default="",
        help="The port (topic) name used for receiving and transmitting facial expressions. "
        "Setting the port name without --set_facial_expressions will only receive the facial expressions",
    )
    parser.add_argument(
        "--mware",
        type=str,
        default=ICUB_DEFAULT_COMMUNICATOR,
        help="The middleware used for communication. "
        "This can be overriden by providing either of the following environment variables "
        "{WRAPYFI_DEFAULT_COMMUNICATOR, WRAPYFI_DEFAULT_MWARE, "
        "ICUB_DEFAULT_COMMUNICATOR, ICUB_DEFAULT_MWARE}. Defaults to 'yarp'",
        choices=MiddlewareCommunicator.get_communicators(),
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    assert not (
        args.headless and (args.set_facial_expressions or args.set_head_eye_coordinates)
    ), "setters require a CV2 window for capturing keystrokes. Disable --set-... for running in headless mode"
    controller = ICub(**vars(args))
    controller.runModule()



import logging

from wrapyfi.connect.clients import Client, Clients


@Clients.register("MMO", "fallback")
class FallbackClient(Client):

    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: str = "",
        missing_middleware_object: str = "",
        **kwargs,
    ):
        logging.warning(
            f"Fallback client employed due to missing middleware or object type: "
            f"{missing_middleware_object}"
        )
        Client.__init__(self, name, in_topic, carrier=carrier, **kwargs)
        self.missing_middleware_object = missing_middleware_object

    def establish(self, repeats: int = -1, **kwargs):
        return None

    def request(self, *args, **kwargs):
        return None

    def _request(self, *args, **kwargs):
        return None

    def _await_reply(self):
        return None

    def close(self):
        return None



"""
This example shows how to use the MiddlewareCommunicator to send and receive messages. It can be used to test the
functionality of the middleware using the PUB/SUB pattern and the REQ/REP pattern. The example can be run on a single
machine or on multiple machines. In this example (as with all other examples), the communication middleware is selected
using the ``--mware`` argument. The default is ZeroMQ, but YARP, ROS, and ROS 2 are also supported.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)

Run:
    # Alternative 1: PUB/SUB mode
        # On machine 1 (or process 1): PUB/SUB mode - Publisher waits for keyboard input and transmits message

        ``python3 hello_world.py --publish --mware zeromq``

        # On machine 2 (or process 2): PUB/SUB mode - Listener waits for message and prints the received object

        ``python3 hello_world.py --listen --mware zeromq``

    # Alternative 2: REQ/REP mode
        # On machine 1 (or process 1): REQ/REP mode - Replier waits for a message, sends a reply, and prints the received object

        ``python3 hello_world.py --reply --mware zeromq``

        # On machine 2 (or process 2): REQ/REP mode - Requester sends a predefined message and waits for a reply

        ``python3 hello_world.py --request --mware zeromq``


"""

import argparse

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class HelloWorld(MiddlewareCommunicator):

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "HelloWorld",
        "/hello/my_message",
        carrier="tcp",
        should_wait=True,
    )
    def send_message(self, arg_from_requester="", mware=None):
        """
        Exchange messages and mirror user input.
        """
        msg = input("Type your message: ")
        obj = {"message": msg, "message_from_requester": arg_from_requester}
        return (obj,)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--publish",
        dest="mode",
        action="store_const",
        const="publish",
        default="listen",
        help="Publish mode",
    )
    parser.add_argument(
        "--listen",
        dest="mode",
        action="store_const",
        const="listen",
        default="listen",
        help="Listen mode (default)",
    )
    parser.add_argument(
        "--request",
        dest="mode",
        action="store_const",
        const="request",
        default="listen",
        help="Request mode",
    )
    parser.add_argument(
        "--reply",
        dest="mode",
        action="store_const",
        const="reply",
        default="listen",
        help="Reply mode",
    )
    parser.add_argument(
        "--mware",
        type=str,
        default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission",
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    hello_world = HelloWorld()
    hello_world.activate_communication(HelloWorld.send_message, mode=args.mode)

    while True:
        (my_message,) = hello_world.send_message(
            arg_from_requester=f"I got this message from the script running in {args.mode} mode",
            mware=args.mware,
        )
        if my_message is not None:
            print("Method result:", my_message)



import time


from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class Test(MiddlewareCommunicator):

    @MiddlewareCommunicator.register(
        "NativeObject", "$mware", "Test", "$topic", should_wait="$should_wait"
    )
    def exchange_object(
        self,
        msg=None,
        mware=DEFAULT_COMMUNICATOR,
        topic="/test/test_native_exchange",
        should_wait=False,
    ):
        ret = {
            "message": msg,
            "set": {"a", 1, None},
            "list": [[[3, [4], 5.677890, 1.2]]],
            "string": "string of characters",
            "2": 2.73211,
            "dict": {
                "other": [
                    None,
                    False,
                    16,
                    4.32,
                ]
            },
        }
        return (ret,)


def test_func(
    queue_buffer,
    mode="listen",
    mware=DEFAULT_COMMUNICATOR,
    topic="/test/test_native_exchange",
    iterations=2,
    should_wait=False,
):
    test = Test()
    test.activate_communication(test.exchange_object, mode=mode)
    for i in range(iterations):
        my_message = test.exchange_object(
            msg=f"signal_idx:{i}", mware=mware, topic=topic, should_wait=should_wait
        )
        if my_message is not None:
            print(f"result {mode}:", my_message[0]["message"])
            queue_buffer.put(my_message[0])
        time.sleep(0.5)



import logging
import sys
import json
import queue
from typing import Optional, Any

import numpy as np
import rospy
import std_msgs.msg

from wrapyfi.connect.clients import Client, Clients
from wrapyfi.middlewares.ros import (
    ROSMiddleware,
    ROSNativeObjectService,
    ROSImageService,
)
from wrapyfi.encoders import JsonEncoder, JsonDecodeHook


class ROSClient(Client):
    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: str = "tcp",
        ros_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the client.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for rep/req pattern. Default is 'tcp'
        :param ros_kwargs: dict: Additional kwargs for the ROS middleware
        :param kwargs: dict: Additional kwargs for the client
        """
        if carrier or carrier != "tcp":
            logging.warning(
                "[ROS] ROS does not support other carriers than TCP for REQ/REP pattern. Using TCP."
            )
            carrier = "tcp"
        super().__init__(name, in_topic, carrier=carrier, **kwargs)
        ROSMiddleware.activate(**ros_kwargs or {})

    def close(self):
        """
        Close the client.
        """
        if hasattr(self, "_client") and self._client:
            self._client.close()

    def __del__(self):
        self.close()


@Clients.register("NativeObject", "ros")
class ROSNativeObjectClient(ROSClient):

    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: str = "tcp",
        persistent: bool = True,
        serializer_kwargs: Optional[dict] = None,
        deserializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The NativeObject client using the ROS String message assuming the data is serialized as a JSON string.
        Deserializes the data (including plugins) using the decoder and parses it to a Python object.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for rep/req pattern. Default is 'tcp'
        :param persistent: bool: Whether to keep the service connection alive across multiple service calls. Default is True
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, in_topic, carrier=carrier, **kwargs)
        self._client = None
        self._queue = queue.Queue(maxsize=1)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self.persistent = persistent

    def establish(self):
        """
        Establish the client's connection to the ROS service.
        """
        rospy.wait_for_service(self.in_topic)
        self._client = rospy.ServiceProxy(
            self.in_topic, ROSNativeObjectService, persistent=self.persistent
        )
        if self.persistent:
            self.established = True

    def request(self, *args, **kwargs) -> Any:
        """
        Send a request to the ROS service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        :return: Any: The response from the ROS service
        """
        if not self.established:
            self.establish()
        try:
            self._request(*args, **kwargs)
        except rospy.ServiceException as e:
            logging.error("[ROS] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the ROS service.

        :param args: tuple: Positional arguments to send in the request.
        :param kwargs: dict: Keyword arguments to send in the request.
        """
        args_str = json.dumps(
            [args, kwargs],
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            serializer_kwrags=self._serializer_kwargs,
        )
        args_msg = std_msgs.msg.String()
        args_msg.data = args_str

        msg = self._client(args_msg)
        obj = json.loads(
            msg.data, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs
        )
        self._queue.put(obj, block=False)

    def _await_reply(self) -> Any:
        """
        Wait for and return the reply from the ROS service.

        :return: Any: The response from the ROS service
        """
        try:
            reply = self._queue.get(block=True)
            return reply
        except queue.Full:
            logging.warning(
                f"[ROS] Discarding data because queue is full. "
                f"This happened due to bad synchronization in {self.__class__.__name__}"
            )
            return None


@Clients.register("Image", "ros")
class ROSImageClient(ROSClient):

    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: str = "tcp",
        width: int = -1,
        height: int = -1,
        persistent: bool = True,
        rgb: bool = True,
        fp: bool = False,
        serializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The Image client using the ROS Image message parsed to a numpy array.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for rep/req pattern. Default is 'tcp'
        :param width: int: The width of the image. Default is -1
        :param height: int: The height of the image. Default is -1
        :param rgb: bool: Whether the image is RGB. Default is True
        :param fp: bool: Whether to utilize floating-point precision. Default is False
        :param persistent: bool: Whether to keep the service connection alive across multiple service calls. Default is True
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        if "jpg" in kwargs:
            logging.warning(
                "[ROS] ROS currently does not support JPG encoding in REQ/REP. Using raw image."
            )
            kwargs.pop("jpg")
        super().__init__(name, in_topic, carrier=carrier, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp

        if self.fp:
            self._encoding = "32FC3" if self.rgb else "32FC1"
            self._type = np.float32
        else:
            self._encoding = "bgr8" if self.rgb else "mono8"
            self._type = np.uint8
        self._pixel_bytes = (3 if self.rgb else 1) * np.dtype(self._type).itemsize

        self._client = None
        self._queue = queue.Queue(maxsize=1)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

        self.persistent = persistent

    def establish(self):
        """
        Establish the client's connection to the ROS service.
        """
        rospy.wait_for_service(self.in_topic)
        self._client = rospy.ServiceProxy(
            self.in_topic, ROSImageService, persistent=self.persistent
        )
        if self.persistent:
            self.established = True

    def request(self, *args, **kwargs):
        """
        Send a request to the ROS service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        :return: np.array: The received image from the ROS service
        """
        if not self.established:
            self.establish()
        try:
            self._request(*args, **kwargs)
        except rospy.ServiceException as e:
            logging.error("[ROS] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the ROS service

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        """
        args_str = json.dumps(
            [args, kwargs],
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            serializer_kwrags=self._serializer_kwargs,
        )
        args_msg = std_msgs.msg.String()
        args_msg.data = args_str
        msg = self._client(args_msg)
        self._queue.put(
            (msg.height, msg.width, msg.encoding, msg.is_bigendian, msg.data),
            block=False,
        )

    def _await_reply(self):
        """
        Wait for and return the reply from the ROS service.

        :return: np.array: The received image from the ROS service
        """
        try:
            height, width, encoding, is_bigendian, data = self._queue.get(block=True)
            if encoding != self._encoding:
                raise ValueError("Incorrect encoding for listener")
            if (
                0 < self.width != width
                or 0 < self.height != height
                or len(data) != height * width * self._pixel_bytes
            ):
                raise ValueError("Incorrect image shape for listener")
            img = np.frombuffer(
                data,
                dtype=np.dtype(self._type).newbyteorder(">" if is_bigendian else "<"),
            ).reshape((height, width, -1))
            if img.shape[2] == 1:
                img = img.squeeze(axis=2)
            return img
        except queue.Full:
            logging.warning(
                f"[ROS] Discarding data because queue is full. "
                f"This happened due to bad synchronization in {self.__name__}"
            )
            return None


@Clients.register("AudioChunk", "ros")
class ROSAudioChunkClient(ROSClient):

    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: str = "tcp",
        persistent: bool = True,
        channels: int = 1,
        rate: int = 44100,
        chunk: int = -1,
        serializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The AudioChunk client using the ROS Audio message parsed to a numpy array.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for rep/req pattern. Default is 'tcp'
        :param channels: int: Number of audio channels. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: The size of audio chunks. Default is -1
        :param persistent: bool: Whether to keep the service connection alive across multiple service calls. Default is True
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        super().__init__(name, in_topic, carrier=carrier, **kwargs)
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._client = self._req_msg = None
        self._queue = queue.Queue(maxsize=1)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

        self.persistent = persistent

    def establish(self):
        """
        Establish the client's connection to the ROS service.
        """
        try:
            from wrapyfi_ros_interfaces.srv import ROSAudioService
        except ImportError:
            import wrapyfi

            logging.error(
                "[ROS] Could not import ROSAudioService. "
                "Make sure the ROS services in wrapyfi_extensions/wrapyfi_ros_interfaces are compiled. "
                "Refer to the documentation for more information: \n"
                + wrapyfi.__doc__
                + "ros_interfaces_lnk.html"
            )
            sys.exit(1)
        self._client = rospy.ServiceProxy(
            self.in_topic, ROSAudioService, persistent=self.persistent
        )
        self._req_msg = ROSAudioService._request_class()
        if self.persistent:
            self.established = True

    def request(self, *args, **kwargs):
        """
        Send a request to the ROS service

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        :return: Tuple[np.array, int]: The received audio chunk and rate from the ROS service
        """
        if not self.established:
            self.establish()
        try:
            self._request(*args, **kwargs)
        except rospy.ServiceException as e:
            logging.error("[ROS] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the ROS service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        """
        args_str = json.dumps(
            [args, kwargs],
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            serializer_kwrags=self._serializer_kwargs,
        )
        args_msg = self._req_msg
        args_msg.request = args_str
        msg = self._client(args_msg).response
        self._queue.put(
            (
                msg.chunk_size,
                msg.channels,
                msg.sample_rate,
                msg.encoding,
                msg.is_bigendian,
                msg.data,
            ),
            block=False,
        )

    def _await_reply(self):
        """
        Wait for and return the reply from the ROS service.

        :return: Tuple[np.array, int]: The received audio chunk and rate from the ROS service
        """
        try:
            chunk, channels, rate, encoding, is_bigendian, data = self._queue.get(
                block=True
            )
            if 0 < self.rate != rate:
                raise ValueError("Incorrect audio rate for client")
            if encoding not in ["S16LE", "S16BE"]:
                raise ValueError("Incorrect encoding for client")
            if (
                0 < self.chunk != chunk
                or self.channels != channels
                or len(data) != chunk * channels * 4
            ):
                raise ValueError("Incorrect audio shape for client")
            aud = np.frombuffer(
                data,
                dtype=np.dtype(np.float32).newbyteorder(">" if is_bigendian else "<"),
            ).reshape((chunk, channels))
            # aud = aud / 32767.0
            return aud, rate
        except queue.Full:
            logging.warning(
                f"[ROS] Discarding data because queue is full. "
                f"This happened due to bad synchronization in {self.__name__}"
            )
            return None, self.rate



"""
A message publisher and listener for native Python objects and Astropy Tables (external plugin).

This script demonstrates the capability to transmit native Python objects and Astropy Tables using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines. By modifying the
``WRAPYFI_PLUGINS_PATH`` environment variable to include the current directory, which contains the
plugins directory with ``astropy_tables.py`` file, the ``AstropyTable`` message can be used to transmit Astropy Tables.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a nested dummy Python object with native objects and Astropy Tables
    - Applying the PUB/SUB pattern with mirroring
    - Using an external plugin to handle Astropy Tables

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - Astropy: Used for creating and handling astronomical tables

    Install using pip:
        ``pip install astropy``

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits message
        ``python3 astropy_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and prints the entire dummy object
        ``python3 astropy_example.py --mode listen``
"""

import os
import argparse

try:
    from astropy.table import Table
except ImportError:
    print("Install Astropy before running this script.")

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR

# Modifying the WRAPYFI_PLUGINS_PATH environment variable to include the plugins directory
script_dir = os.path.dirname(os.path.realpath(__file__))
if "WRAPYFI_PLUGINS_PATH" in os.environ:
    os.environ["WRAPYFI_PLUGINS_PATH"] += os.pathsep + script_dir
else:
    os.environ["WRAPYFI_PLUGINS_PATH"] = script_dir


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "Notifier",
        "/notify/test_astropy_exchange",
        carrier="tcp",
        should_wait=True,
    )
    def exchange_object(self, mware=None):
        """
        Exchange messages with Astropy Tables and other native Python objects.
        """
        msg = input("Type your message: ")

        # Creating an example Astropy Table
        t = Table()
        t["name"] = ["source 1", "source 2", "source 3"]
        t["flux"] = [1.2, 2.2, 3.1]

        ret = {
            "message": msg,
            "astropy_table": t,
        }
        return (ret,)


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="A message publisher and listener for native Python objects and Astropy Tables."
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="publish",
        choices={"publish", "listen"},
        help="The transmission mode",
    )
    parser.add_argument(
        "--mware",
        type=str,
        default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission",
    )
    return parser.parse_args()


def main(args):
    """
    Main function to initiate Notifier class and communication.
    """
    notifier = Notifier()
    notifier.activate_communication(Notifier.exchange_object, mode=args.mode)

    while True:
        (msg_object,) = notifier.exchange_object(mware=args.mware)
        print("Method result:", msg_object)


if __name__ == "__main__":
    args = parse_args()
    main(args)



import yaml
import logging
from typing import Union, Optional

from wrapyfi.utils import SingletonOptimized


class ConfigManager(metaclass=SingletonOptimized):
    """
    The configuration manager is a singleton which is invoked once throughout the runtime.
    """

    def __init__(self, config: Optional[Union[dict, str]], **kwargs):
        """
        Initializing the ConfigManager. The configuration can be provided as a yaml file name or as a dictionary.

        :param config: Union[dict, str]: Loads a yaml configuration file when "str" type provided. Directly set when "dict" type provided
        """
        if isinstance(config, str):
            self.config = self.__loadfile(config)
            logging.info(f"Loaded Wrapyfi configuration: {self.config}")
        elif isinstance(config, dict):
            self.config = config
            logging.info(f"Loaded Wrapyfi configuration: {self.config}")
        else:
            self.config = []

    @staticmethod
    def __loadfile(filename: str):
        """
        Loading the yaml configuration file.

        :param filename: str: The file name
        :return: dict: The configuration object
        """
        with open(filename, "r") as fp:
            config = yaml.safe_load(fp)
        return config

    def __writefile(self, filename: str):
        """
        Saving the yaml configuration file.

        :param filename: str: The file name
        """
        with open(filename, "w") as fp:
            yaml.safe_dump(self.config, fp)



import unittest
import multiprocessing
from multiprocessing import Queue
import queue


class ZeroMQTestMiddleware(unittest.TestCase):
    MWARE = "zeromq"

    def test_publish_listen(self):
        """
        Test the publish and listen functionality of the middleware. This verifies that the middleware can send and
        receive messages using the PUB/SUB pattern.
        """
        import wrapyfi.tests.tools.class_test as class_test

        # importlib.reload(wrapyfi)
        # importlib.reload(class_test)
        test_func = class_test.test_func
        Test = class_test.Test
        Test.close_all_instances()

        if self.MWARE not in class_test.Test.get_communicators():
            self.skipTest(f"{self.MWARE} not installed")

        listen_queue = Queue(maxsize=10)
        test_lsn = multiprocessing.Process(
            target=test_func,
            args=(listen_queue,),
            kwargs={
                "mode": "listen",
                "mware": self.MWARE,
                "iterations": 10,
                "should_wait": True,
            },
        )
        # test_lsn.daemon = True

        publish_queue = Queue(maxsize=10)
        test_pub = multiprocessing.Process(
            target=test_func,
            args=(publish_queue,),
            kwargs={
                "mode": "publish",
                "mware": self.MWARE,
                "iterations": 10,
                "should_wait": True,
            },
        )
        # test_pub.daemon = True

        test_lsn.start()
        test_pub.start()
        test_lsn.join()
        test_pub.join()
        for i in range(10):
            try:
                self.assertDictEqual(
                    listen_queue.get(timeout=3), publish_queue.get(timeout=3)
                )
            except queue.Empty:
                self.assertEqual(i, 9)


class ROS2TestMiddleware(ZeroMQTestMiddleware):
    """
    Test the ROS 2 wrapper. This test class inherits from the ZeroMQ test class, so all tests from the ZeroMQ test class
    are also run for the ROS 2 wrapper.
    """

    MWARE = "ros2"


class YarpTestMiddleware(ZeroMQTestMiddleware):
    """
    Test the YARP wrapper. This test class inherits from the ZeroMQ test class, so all tests from the ZeroMQ test class
    are also run for the YARP wrapper.
    """

    MWARE = "yarp"


class ROSTestMiddleware(ZeroMQTestMiddleware):
    """
    Test the ROS wrapper. This test class inherits from the ZeroMQ test class, so all tests from the ZeroMQ test class
    are also run for the ROS wrapper.
    """

    MWARE = "ros"


if __name__ == "__main__":
    unittest.main()



import logging
import os
from glob import glob
from pathlib import Path


from wrapyfi.utils import SingletonOptimized, dynamic_module_import


class ListenerWatchDog(metaclass=SingletonOptimized):
    """
    A watchdog that scans for listeners and removes them from the ring if they are not established.
    """

    def __init__(self, repeats: int = 10, inner_repeats: int = 10):
        """
        Initialize the ListenerWatchDog.

        :param repeats: int: The number of times to repeat the scan
        param inner_repeats: int: The number of times to repeat the scan for each listener
        """
        self.repeats = repeats
        self.inner_repeats = inner_repeats
        self.listener_ring = []

    def add_listener(self, listener):
        """
        Add a listener to the ring.

        :param listener: Listener: The listener to add
        """
        self.listener_ring.append(listener)

    def remove_listener(self, listener):
        """
        Remove a listener from the ring.

        :param listener: Listener: The listener to remove
        """
        self.listener_ring.remove(listener)

    def scan(self):
        """
        Scan for listeners and remove them from the ring if they are established.
        """
        repeats = self.repeats
        while self.listener_ring and (repeats > 0 | repeats <= -1):
            repeats -= 1
            for listener in self.listener_ring:
                found_listener = listener.establish(repeats=self.inner_repeats)
                if found_listener:
                    self.listener_ring.remove(listener)


class Listeners(object):
    """
    A class that holds all listeners and their corresponding middleware communicators.
    """

    registry = {}
    mwares = set()

    @classmethod
    def register(cls, data_type: str, communicator: str):
        """
        Register a listener with the given data type and middleware communicator.

        :param data_type: str: The data type to register the listener for e.g., "NativeObject", "Image", "AudioChunk", etc.
        :param communicator: str: The middleware communicator to register the listener for e.g., "ros", "ros2", "yarp", "zeromq", etc.
        """

        def decorator(cls_):
            cls.registry[data_type + ":" + communicator] = cls_
            cls.mwares.add(communicator)
            return cls_

        return decorator

    @staticmethod
    def scan():
        """
        Scan for listeners and add them to the registry.
        """
        # modules = glob(
        #     os.path.join(os.path.dirname(__file__), "..", "listeners", "*.py"),
        #     recursive=True,
        # )
        # modules = [
        #     "wrapyfi.listeners."
        #     + module.replace(os.path.dirname(__file__) + "/../listeners/", "")
        #     for module in modules
        # ]
        # dynamic_module_import(modules, globals())
        base_dir = Path(__file__).parent.parent / "listeners"
        modules = glob(str(base_dir / "*.py"), recursive=True)

        modules = [
            "wrapyfi.listeners." + Path(module).relative_to(base_dir).as_posix()
            for module in modules
        ]
        dynamic_module_import(modules, globals())


class Listener(object):
    """
    A base class for listeners.
    """

    def __init__(
        self,
        name: str,
        in_topic: str,
        carrier: str = "",
        should_wait: bool = True,
        **kwargs,
    ):
        """
        Initialize the Listener.

        :param name: str: The name of the listener
        :param in_topic: str: The topic to listen to
        :param carrier: str: The middleware carrier to use
        :param should_wait: bool: Whether to wait for the listener to be established or not
        """
        self.__name__ = name
        self.in_topic = in_topic
        self.carrier = carrier
        self.should_wait = should_wait
        self.established = False

    def check_establishment(self, established: bool):
        """
        Check if the listener is established or not.

        :param established: bool: Whether the listener is established or not
        :return: bool: Whether the listener is established or not
        """
        if established:
            self.established = True
            if not self.should_wait:
                ListenerWatchDog().remove_listener(self)
        elif not self.should_wait:
            ListenerWatchDog().scan()
            if self in ListenerWatchDog().listener_ring:
                established = False
            else:
                established = True
        return established

    def establish(self, repeats: int = -1, **kwargs):
        """
        Establish the listener.
        """
        raise NotImplementedError

    def listen(self):
        """
        Listen for incoming data.
        """
        raise NotImplementedError

    def close(self):
        """
        Close the connection.
        """
        raise NotImplementedError



import argparse

import cv2

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR
from wrapyfi.config.manager import ConfigManager


class ExperimentController(MiddlewareCommunicator):
    WEBCAM_HEIGHT = 240
    WEBCAM_WIDTH = 320
    ICUB_CAM_HEIGHT = 240
    ICUB_CAM_WIDTH = 320
    SIXDREPNET_CAM_HEIGHT = 240
    SIXDREPNET_CAM_WIDTH = 320

    def __init__(self, **kwargs):
        super(ExperimentController, self).__init__()

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_sixdrepnet_mware",
        "ExperimentController",
        "/control_interface/orientation_sixdrepnet",
        should_wait=False,
    )
    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_waveshareimu_mware",
        "ExperimentController",
        "/control_interface/orientation_waveshareimu",
        should_wait=False,
    )
    def listen_orientation_sixdrepnet_waveshareimu(
        self, _sixdrepnet_mware="yarp", _waveshareimu_mware="yarp"
    ):
        return (
            None,
            None,
        )

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ExperimentController",
        "/control_interface/orientation_icub",
        should_wait=False,
    )
    def publish_orientation_icub(self, obj, _mware="yarp"):
        return (obj,)

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ExperimentController",
        "/control_interface/gaze_pupil",
        should_wait=False,
    )
    def listen_gaze_pupil(self, _mware="zeromq"):
        return (None,)

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ExperimentController",
        "/control_interface/gaze_icub",
        should_wait=False,
    )
    def publish_gaze_icub(self, obj, _mware="yarp"):
        return (obj,)

    @MiddlewareCommunicator.register(
        "Image",
        "$_mware",
        "ExperimentController",
        "/control_interface/image_webcam",
        width="$_width",
        height="$_height",
        rgb=True,
        fp=False,
        should_wait=False,
        jpg=True,
    )
    def listen_image_webcam(
        self, _mware="ros2", _width=WEBCAM_WIDTH, _height=WEBCAM_HEIGHT
    ):
        return (None,)

    @MiddlewareCommunicator.register(
        "Image",
        "$_mware",
        "ExperimentController",
        "/control_interface/image_sixdrepnet",
        width="$_width",
        height="$_height",
        rgb=True,
        fp=False,
        should_wait=False,
    )
    def publish_sixdrepnet_cam(
        self,
        img,
        _mware="ros2",
        _width=SIXDREPNET_CAM_WIDTH,
        _height=SIXDREPNET_CAM_HEIGHT,
    ):
        # modify size of image
        if img is not None:
            img = cv2.resize(
                img, (self.SIXDREPNET_CAM_WIDTH, self.SIXDREPNET_CAM_HEIGHT)
            )
        return (img,)

    def priority_control_sources(
        self, orientation_sixdrepnet, orientation_imu, control_sources
    ):
        if control_sources[0] == "vision":
            if orientation_sixdrepnet is not None:
                # print("Orientation 6DRepNet: ", orientation_sixdrepnet)
                self.publish_orientation_icub(orientation_sixdrepnet)
            elif orientation_imu is not None and "imu" in control_sources:
                # print("Orientation IMU: ", orientation_imu)
                self.publish_orientation_icub(orientation_imu)
        elif control_sources[0] == "imu":
            if orientation_imu is not None:
                # print("Orientation IMU: ", orientation_imu)
                self.publish_orientation_icub(orientation_imu)
            elif orientation_sixdrepnet is not None and "vision" in control_sources:
                # print("Orientation 6DRepNet: ", orientation_sixdrepnet)
                self.publish_orientation_icub(orientation_sixdrepnet)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--wrapyfi_cfg",
        help="File to load Wrapyfi configs for running instance. "
        "Choose one of the configs available in "
        "./wrapyfi_configs/gaze_mirroring_multisensor. "
        "for each running instance. All configs with a prefix of COMP must "
        "have corresponding instances running. OPT prefixed config is optional "
        "(only when using a vision model) executes the script on a machine connected "
        "to the camera (or has access to the camera topic).",
        type=str,
    )
    parser.add_argument(
        "--control_sources",
        help="Control sources to use for the experiment. The order of sources indicates the priority. "
        "For example, if vision is the first source, then the vision source will "
        "be used for control. If vision is not available, then the IMU source will be used. "
        "If one source is provided, then the experiment will run with that source. ",
        type=str,
        default=["vision", "imu"],
        nargs="+",
        choices=["vision", "imu"],
    )
    parser.add_argument(
        "--enable_gaze",
        help="Enable the gaze (eye movement) control. If not enabled, "
        "then the gaze control will not be used. ",
        action="store_true",
        default=False,
    )

    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    if args.wrapyfi_cfg:
        print("Wrapyfi config loading from ", args.wrapyfi_cfg)
        ConfigManager(args.wrapyfi_cfg)

    ec = ExperimentController()
    image_cam = None
    while True:
        if "vision" in args.control_sources:
            (image_webcam,) = ec.listen_image_webcam()
            (image_webcam_linked,) = ec.publish_sixdrepnet_cam(image_cam)
            if image_webcam is None:
                image_webcam = image_webcam_linked
            image_cam = image_webcam
            if image_webcam is not None:
                cv2.imshow("Webcam image", image_webcam)
                cv2.waitKey(1)

        orientation_sixdrepnet, orientation_imu = (
            ec.listen_orientation_sixdrepnet_waveshareimu()
        )
        ec.priority_control_sources(
            orientation_sixdrepnet, orientation_imu, args.control_sources
        )

        if args.enable_gaze:
            (gaze_pupil,) = ec.listen_gaze_pupil()
            if gaze_pupil is not None:
                # print("Gaze Pupil: ", gaze_pupil)
                ec.publish_gaze_icub(gaze_pupil)
